# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import itertools
import logging
import time
from collections import defaultdict
from typing import Dict, List

from ..... import oscar as mo
from .....core import ChunkGraph, Chunk
from .....core.operand import Fuse, Fetch
from .....metrics import Metrics
from .....utils import get_chunk_params
from .....typing import BandType, TileableType
from ....meta import MetaAPI, WorkerMetaAPI
from ....scheduling import SchedulingAPI
from ....subtask import Subtask, SubtaskGraph, SubtaskResult, SubtaskStatus
from ....task.core import Task, TaskResult, TaskStatus
from ..api import ExecutionChunkResult

logger = logging.getLogger(__name__)


class TaskStageProcessor:
    def __init__(
        self,
        stage_id: str,
        task: Task,
        chunk_graph: ChunkGraph,
        subtask_graph: SubtaskGraph,
        bands: List[BandType],
        tile_context: Dict[TileableType, TileableType],
        scheduling_api: SchedulingAPI,
        meta_api: MetaAPI,
    ):
        self.stage_id = stage_id
        self.task = task
        self.chunk_graph = chunk_graph
        self.subtask_graph = subtask_graph
        self._bands = bands
        self._tile_context = tile_context

        # APIs
        self._scheduling_api = scheduling_api
        self._meta_api = meta_api

        # gen subtask_id to subtask
        self.subtask_id_to_subtask = {
            subtask.subtask_id: subtask for subtask in subtask_graph
        }
        self._subtask_to_bands: Dict[Subtask, BandType] = dict()
        self.subtask_snapshots: Dict[Subtask, SubtaskResult] = dict()
        self.subtask_results: Dict[Subtask, SubtaskResult] = dict()
        self._submitted_subtask_ids = set()

        # All subtask IDs whose input chunk reference count is reduced.
        self.decref_subtask = set()

        self._band_manager: Dict[BandType, mo.ActorRef] = dict()

        # result
        self.result = TaskResult(
            task.task_id,
            task.session_id,
            self.stage_id,
            status=TaskStatus.pending,
            start_time=time.time(),
        )
        # status
        self._done = asyncio.Event()
        self._cancelled = asyncio.Event()

        # add metrics
        self._stage_execution_time = Metrics.gauge(
            "mars.stage_execution_time_secs",
            "Time consuming in seconds to execute a stage",
            ("session_id", "task_id", "stage_id"),
        )

    def is_cancelled(self):
        return self._cancelled.is_set()

    async def _schedule_subtasks(self, subtasks: List[Subtask]):
        subtasks = [
            subtask
            for subtask in subtasks
            if subtask.subtask_id not in self._submitted_subtask_ids
        ]
        if not subtasks:
            return
        self._submitted_subtask_ids.update(subtask.subtask_id for subtask in subtasks)
        return await self._scheduling_api.add_subtasks(
            subtasks, [subtask.priority for subtask in subtasks]
        )

    async def _get_stage_result(self):
        chunks = []
        get_meta = []
        results_chunks = self.chunk_graph.result_chunks
        for chunk in results_chunks:
            if isinstance(chunk.op, Fetch):
                continue
            chunks.append(chunk)
            if isinstance(chunk.op, Fuse):
                chunk = chunk.chunk
            get_meta.append(
                self._meta_api.get_chunk_meta.delay(
                    chunk.key,
                    # only fetch bands from supervisor meta
                    fields=["bands"],
                )
            )
        metas = await self._meta_api.get_chunk_meta.batch(*get_meta)
        execution_chunk_results = {
            chunk: ExecutionChunkResult(meta=meta, context=None)
            for chunk, meta in zip(chunks, metas)
        }
        await self._update_result_meta(execution_chunk_results)
        return execution_chunk_results

    def _schedule_done(self):
        self._done.set()

    async def set_subtask_result(self, result: SubtaskResult, band: BandType = None):
        assert result.status.is_done
        subtask = self.subtask_id_to_subtask[result.subtask_id]
        #  update subtask_results in `TaskProcessorActor.set_subtask_result`
        self._submitted_subtask_ids.difference_update([result.subtask_id])

        all_done = len(self.subtask_results) == len(self.subtask_graph)
        error_or_cancelled = result.status in (
            SubtaskStatus.errored,
            SubtaskStatus.cancelled,
        )

        if all_done or error_or_cancelled:
            # tell scheduling to finish subtasks
            await self._scheduling_api.finish_subtasks(
                [result.subtask_id], bands=[band], schedule_next=not error_or_cancelled
            )
            if self.result.status != TaskStatus.terminated:
                self.result = TaskResult(
                    self.task.task_id,
                    self.task.session_id,
                    self.stage_id,
                    start_time=self.result.start_time,
                    end_time=time.time(),
                    status=TaskStatus.terminated,
                    error=result.error,
                    traceback=result.traceback,
                )
                if not all_done and error_or_cancelled:
                    if result.status == SubtaskStatus.errored:
                        logger.exception(
                            "Subtask %s errored",
                            subtask.subtask_id,
                            exc_info=(
                                type(result.error),
                                result.error,
                                result.traceback,
                            ),
                        )
                    if result.status == SubtaskStatus.cancelled:  # pragma: no cover
                        logger.warning(
                            "Subtask %s from band %s canceled.",
                            subtask.subtask_id,
                            band,
                        )
                    logger.info(
                        "Start to cancel stage %s of task %s.", self.stage_id, self.task
                    )
                    # if error or cancel, cancel all submitted subtasks
                    await self._scheduling_api.cancel_subtasks(
                        list(self._submitted_subtask_ids)
                    )
                self._schedule_done()
                cost_time_secs = self.result.end_time - self.result.start_time
                logger.info(
                    "Time consuming to execute a stage is %ss with "
                    "session id %s, task id %s, stage id %s",
                    cost_time_secs,
                    self.result.session_id,
                    self.result.task_id,
                    self.result.stage_id,
                )
                self._stage_execution_time.record(
                    cost_time_secs,
                    {
                        "session_id": self.result.session_id,
                        "task_id": self.result.task_id,
                        "stage_id": self.result.stage_id,
                    },
                )
        else:
            # not terminated, push success subtasks to queue if they are ready
            to_schedule_subtasks = []
            for succ_subtask in self.subtask_graph.successors(subtask):
                if succ_subtask in self.subtask_results:  # pragma: no cover
                    continue
                pred_subtasks = self.subtask_graph.predecessors(succ_subtask)
                if all(
                    pred_subtask in self.subtask_results
                    for pred_subtask in pred_subtasks
                ):
                    # all predecessors finished
                    to_schedule_subtasks.append(succ_subtask)
            await self._schedule_subtasks(to_schedule_subtasks)
            await self._scheduling_api.finish_subtasks(
                [result.subtask_id], bands=[band]
            )

    async def run(self):
        if len(self.subtask_graph) == 0:
            # no subtask to schedule, set status to done
            self._schedule_done()
            self.result.status = TaskStatus.terminated
            return {}

        # schedule independent subtasks
        indep_subtasks = list(self.subtask_graph.iter_indep())
        await self._schedule_subtasks(indep_subtasks)

        # wait for completion
        await self._done.wait()
        if self.error_or_cancelled():
            if self.result.error is not None:
                raise self.result.error.with_traceback(self.result.traceback)
            else:
                raise asyncio.CancelledError()
        return await self._get_stage_result()

    async def cancel(self):
        logger.info("Start to cancel stage %s of task %s.", self.stage_id, self.task)
        if self._done.is_set():  # pragma: no cover
            # already finished, ignore cancel
            return
        self._cancelled.set()
        # cancel running subtasks
        await self._scheduling_api.cancel_subtasks(list(self._submitted_subtask_ids))
        self._done.set()

    def error_or_cancelled(self) -> bool:
        if self.result.error is not None:
            return True
        if self.is_cancelled():
            return True
        return False

    async def _update_result_meta(
        self, chunk_to_result: Dict[Chunk, ExecutionChunkResult]
    ):
        session_id = self.task.session_id
        tile_context = self._tile_context

        update_meta_chunks = chunk_to_result.keys() - set(
            itertools.chain.from_iterable(
                (c.data for c in tiled_tileable.chunks)
                for tiled_tileable in tile_context.values()
            )
        )

        worker_meta_api_to_chunk_delays = defaultdict(dict)
        for c in update_meta_chunks:
            address = chunk_to_result[c].meta["bands"][0][0]
            meta_api = await WorkerMetaAPI.create(session_id, address)
            call = meta_api.get_chunk_meta.delay(
                c.key, fields=list(get_chunk_params(c).keys())
            )
            worker_meta_api_to_chunk_delays[meta_api][c] = call
        for tileable in tile_context.values():
            chunks = [c.data for c in tileable.chunks]
            for c, params_fields in zip(chunks, self._get_params_fields(tileable)):
                address = chunk_to_result[c].meta["bands"][0][0]
                meta_api = await WorkerMetaAPI.create(session_id, address)
                call = meta_api.get_chunk_meta.delay(c.key, fields=params_fields)
                worker_meta_api_to_chunk_delays[meta_api][c] = call
        coros = []
        for worker_meta_api, chunk_delays in worker_meta_api_to_chunk_delays.items():
            coros.append(worker_meta_api.get_chunk_meta.batch(*chunk_delays.values()))
        worker_metas = await asyncio.gather(*coros)
        for chunk_delays, metas in zip(
            worker_meta_api_to_chunk_delays.values(), worker_metas
        ):
            for c, meta in zip(chunk_delays, metas):
                chunk_to_result[c].meta = meta

    @classmethod
    def _get_params_fields(cls, tileable: TileableType):
        from .....dataframe.core import DATAFRAME_TYPE, SERIES_TYPE
        from .....tensor.core import TENSOR_TYPE

        params_fields = []
        fields = get_chunk_params(tileable.chunks[0])
        if isinstance(tileable, DATAFRAME_TYPE):
            for c in tileable.chunks:
                cur_fields = set(fields)
                if c.index[1] > 0:
                    # skip fetch index_value for i >= 1 on column axis
                    cur_fields.remove("index_value")
                if c.index[0] > 0:
                    # skip fetch dtypes_value for i >= 1 on index axis
                    cur_fields.remove("dtypes_value")
                if c.index[0] > 0 and c.index[1] > 0:
                    # fetch shape only for i == 0 on index or column axis
                    cur_fields.remove("shape")
                params_fields.append(list(cur_fields))
        elif isinstance(tileable, SERIES_TYPE):
            for c in tileable.chunks:
                cur_fields = set(fields)
                if c.index[0] > 0:
                    # skip fetch name and dtype for i >= 1
                    cur_fields.remove("name")
                    cur_fields.remove("dtype")
                params_fields.append(list(cur_fields))
        elif isinstance(tileable, TENSOR_TYPE):
            for i, c in enumerate(tileable.chunks):
                cur_fields = set(fields)
                if c.ndim > 1 and all(j > 0 for j in c.index):
                    cur_fields.remove("shape")
                if i > 0:
                    cur_fields.remove("dtype")
                    cur_fields.remove("order")
                params_fields.append(list(cur_fields))
        else:
            for _ in tileable.chunks:
                params_fields.append(list(fields))
        return params_fields



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, recursive_tile
from ...serialization.serializables import KeyField, AnyField, BoolField, Int32Field
from ...utils import has_unknown_shape, ceildiv
from ..operands import TensorOperand, TensorOperandMixin
from ..datasource import tensor as astensor
from ..array_utils import as_same_device, device
from ..core import Tensor, TENSOR_TYPE
from ..utils import decide_unify_split


class TensorFillDiagonal(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.FILL_DIAGONAL

    _input = KeyField("input")
    _val = AnyField("val")
    _wrap = BoolField("wrap")
    # used for chunk
    _k = Int32Field("k")

    def __init__(self, val=None, wrap=None, k=None, **kw):
        super().__init__(_val=val, _wrap=wrap, _k=k, **kw)

    @property
    def input(self):
        return self._input

    @property
    def val(self):
        return self._val

    @property
    def wrap(self):
        return self._wrap

    @property
    def k(self):
        return self._k

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if len(self._inputs) == 2:
            self._val = self._inputs[1]

    def __call__(self, a, val=None):
        inputs = [a]
        if val is not None:
            inputs.append(val)
        return self.new_tensor(inputs, shape=a.shape, order=a.order)

    @staticmethod
    def _process_val(val, a, wrap):
        """
        given the `val`, `a`, `wrap` which are the arguments in `fill_diagonal`,
        do some preprocess on `val` includes:

        1. calculate the length to fill on diagonal, 2-d and n-d(n > 2)
           as well as that `wrap` is True and `a` is a tall matrix need to be considered.
        2. if val is a Tensor, rechunk it into one chunk.
        """

        from ..datasource import diag
        from ..base import tile

        is_val_tensor = isinstance(val, TENSOR_TYPE)

        if a.ndim == 2:
            if wrap and TensorFillDiagonal._is_tall(a):
                size = sum(
                    diag(sub).shape[0]
                    for sub in TensorFillDiagonal._split_tall_matrix(a)
                )
            else:
                size = diag(a).shape[0]
        else:
            # every dimension has same shape
            size = a.shape[0]

        repeat_method = tile if is_val_tensor else np.tile
        val_size = val.size
        if val_size < size:
            n = ceildiv(size, val_size)
            val = repeat_method(val, n)[:size]
        elif val_size > size:
            val = val[:size]

        if is_val_tensor and val.ndim > 0:
            val = yield from recursive_tile(val)
            val = val.rechunk({0: val.size})

        return (yield from recursive_tile(val)) if is_val_tensor else val

    @staticmethod
    def _gen_val(val, diag_idx, cum_sizes):
        """
        Given a tensor-level `val`, calculate the chunk-level `val`.
        Consider both the cases that `val` could be a tensor or ndarray.

        :param val: tensor-level `val`
        :diag_idx: chunk index on the diagonal direction
        :cum_sizes: accumulative chunk sizes on the diagonal direction
        """
        from .slice import TensorSlice

        if val.ndim == 0:
            if isinstance(val, TENSOR_TYPE):
                return val.chunks[0]
            else:
                return val

        if isinstance(val, TENSOR_TYPE):
            start, stop = cum_sizes[diag_idx], cum_sizes[diag_idx + 1]
            slc = slice(start, stop)
            slc_op = TensorSlice(slices=[slc], dtype=val.dtype)
            return slc_op.new_chunk(
                [val.chunks[0]],
                shape=(stop - start,),
                order=val.order,
                index=(diag_idx,),
            )
        else:
            return val[cum_sizes[diag_idx] : cum_sizes[diag_idx + 1]]

    @classmethod
    def _tile_2d(cls, op, val):
        from ..datasource import diag

        d = yield from recursive_tile(diag(op.input))
        index_to_diag_chunk = {c.inputs[0].index: c for c in d.chunks}
        cum_sizes = [0] + np.cumsum(d.nsplits[0]).tolist()

        out_chunks = []
        for chunk in op.input.chunks:
            if chunk.index not in index_to_diag_chunk:
                out_chunks.append(chunk)
            else:
                diag_chunk = index_to_diag_chunk[chunk.index]
                diag_idx = diag_chunk.index[0]
                input_chunks = [chunk]
                chunk_val = cls._gen_val(val, diag_idx, cum_sizes)
                if len(op.inputs) == 2:
                    input_chunks.append(chunk_val)
                chunk_op = op.copy().reset_key()
                chunk_op._wrap = False
                chunk_op._k = diag_chunk.op.k
                chunk_op._val = chunk_val
                out_chunk = chunk_op.new_chunk(
                    input_chunks,
                    shape=chunk.shape,
                    order=chunk.order,
                    index=chunk.index,
                )
                out_chunks.append(out_chunk)

        out = op.outputs[0]
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            shape=out.shape,
            order=out.order,
            chunks=out_chunks,
            nsplits=op.input.nsplits,
        )

    @classmethod
    def _tile_nd(cls, op, val):
        # if more than 3d, we will rechunk the tensor into square chunk
        # on the diagonal direction
        in_tensor = op.input
        nsplits = np.array(in_tensor.nsplits)
        if not np.issubdtype(nsplits.dtype, np.integer) or not np.all(
            np.diff(nsplits, axis=1) == 0
        ):
            # need rechunk
            nsplit = decide_unify_split(*in_tensor.nsplits)
            in_tensor = yield from recursive_tile(
                in_tensor.rechunk(tuple(nsplit for _ in range(in_tensor.ndim)))
            )
        cum_sizes = [0] + np.cumsum(in_tensor.nsplits[0]).tolist()

        out_chunks = []
        for chunk in in_tensor.chunks:
            if len(set(chunk.index)) == 1:
                # chunk on the diagonal direction
                chunk_op = op.copy().reset_key()
                chunk_op._k = 0
                chunk_inputs = [chunk]
                chunk_val = cls._gen_val(val, chunk.index[0], cum_sizes)
                if len(op.inputs) == 2:
                    chunk_inputs.append(chunk_val)
                chunk_op._val = chunk_val
                out_chunk = chunk_op.new_chunk(
                    chunk_inputs,
                    shape=chunk.shape,
                    order=chunk.order,
                    index=chunk.index,
                )
                out_chunks.append(out_chunk)
            else:
                out_chunks.append(chunk)

        out = op.outputs[0]
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            shape=out.shape,
            order=out.order,
            chunks=out_chunks,
            nsplits=in_tensor.nsplits,
        )

    @classmethod
    def _tile_one_chunk(cls, op, val):
        out = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk_inputs = [op.input.chunks[0]]
        if isinstance(val, TENSOR_TYPE):
            chunk_inputs.append(val.chunks[0])
        chunk = chunk_op.new_chunk(
            chunk_inputs, shape=out.shape, order=out.order, index=(0,) * out.ndim
        )
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            shape=out.shape,
            order=out.order,
            chunks=[chunk],
            nsplits=((s,) for s in out.shape),
        )

    @staticmethod
    def _is_tall(x):
        return x.shape[0] > x.shape[1] + 1

    @staticmethod
    def _split_tall_matrix(a):
        blocksize = a.shape[1] + 1
        n_block = ceildiv(a.shape[0], blocksize)
        return [a[i * blocksize : (i + 1) * blocksize] for i in range(n_block)]

    @classmethod
    def tile(cls, op):
        # input tensor must have no unknown chunk shape
        if has_unknown_shape(*op.inputs):
            yield

        in_tensor = op.input
        is_in_tensor_tall = cls._is_tall(in_tensor)

        if op.val.ndim > 0:
            val = yield from cls._process_val(op.val, in_tensor, op.wrap)
        else:
            val = op.val

        if len(in_tensor.chunks) == 1:
            return cls._tile_one_chunk(op, val)

        if op.input.ndim == 2:
            if op.wrap and is_in_tensor_tall:
                from ..merge import concatenate

                sub_tensors = cls._split_tall_matrix(in_tensor)
                for i, sub_tensor in enumerate(sub_tensors):
                    if val.ndim > 0:
                        sub_val = val[
                            i * sub_tensor.shape[1] : (i + 1) * sub_tensor.shape[1]
                        ]
                    else:
                        sub_val = val
                    fill_diagonal(sub_tensor, sub_val, wrap=False)
                out_tensor = concatenate(sub_tensors)
                return [(yield from recursive_tile(out_tensor))]
            else:
                return (yield from cls._tile_2d(op, val))
        else:
            return (yield from cls._tile_nd(op, val))

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True
        )
        a = inputs[0]
        if len(inputs) == 2:
            val = inputs[1]
        else:
            val = op.val

        with device(device_id):
            if not op.k:
                a = a.copy()
                xp.fill_diagonal(a, val, wrap=op.wrap)
            else:
                assert a.ndim == 2
                k = op.k or 0
                n_rows, n_cols = a.shape
                if k > 0:
                    n_cols -= k
                elif k < 0:
                    n_rows += k
                n = min(n_rows, n_cols)

                # generate indices
                rows, cols = np.diag_indices(n)
                if k > 0:
                    cols = cols.copy()
                    cols += k
                elif k < 0:
                    rows = rows.copy()
                    rows -= k

                a = a.copy()
                a[rows, cols] = val

            ctx[op.outputs[0].key] = a


def fill_diagonal(a, val, wrap=False):
    """Fill the main diagonal of the given tensor of any dimensionality.

    For a tensor `a` with ``a.ndim >= 2``, the diagonal is the list of
    locations with indices ``a[i, ..., i]`` all identical. This function
    modifies the input tensor in-place, it does not return a value.

    Parameters
    ----------
    a : Tensor, at least 2-D.
      Tensor whose diagonal is to be filled, it gets modified in-place.

    val : scalar
      Value to be written on the diagonal, its type must be compatible with
      that of the tensor a.

    wrap : bool
      For tall matrices in NumPy version up to 1.6.2, the
      diagonal "wrapped" after N columns. You can have this behavior
      with this option. This affects only tall matrices.

    See also
    --------
    diag_indices, diag_indices_from

    Notes
    -----

    This functionality can be obtained via `diag_indices`, but internally
    this version uses a much faster implementation that never constructs the
    indices and uses simple slicing.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.zeros((3, 3), int)
    >>> mt.fill_diagonal(a, 5)
    >>> a.execute()
    array([[5, 0, 0],
           [0, 5, 0],
           [0, 0, 5]])

    The same function can operate on a 4-D tensor:

    >>> a = mt.zeros((3, 3, 3, 3), int)
    >>> mt.fill_diagonal(a, 4)

    We only show a few blocks for clarity:

    >>> a[0, 0].execute()
    array([[4, 0, 0],
           [0, 0, 0],
           [0, 0, 0]])
    >>> a[1, 1].execute()
    array([[0, 0, 0],
           [0, 4, 0],
           [0, 0, 0]])
    >>> a[2, 2].execute()
    array([[0, 0, 0],
           [0, 0, 0],
           [0, 0, 4]])

    The wrap option affects only tall matrices:

    >>> # tall matrices no wrap
    >>> a = mt.zeros((5, 3), int)
    >>> mt.fill_diagonal(a, 4)
    >>> a.execute()
    array([[4, 0, 0],
           [0, 4, 0],
           [0, 0, 4],
           [0, 0, 0],
           [0, 0, 0]])

    >>> # tall matrices wrap
    >>> a = mt.zeros((5, 3), int)
    >>> mt.fill_diagonal(a, 4, wrap=True)
    >>> a.execute()
    array([[4, 0, 0],
           [0, 4, 0],
           [0, 0, 4],
           [0, 0, 0],
           [4, 0, 0]])

    >>> # wide matrices
    >>> a = mt.zeros((3, 5), int)
    >>> mt.fill_diagonal(a, 4, wrap=True)
    >>> a.execute()
    array([[4, 0, 0, 0, 0],
           [0, 4, 0, 0, 0],
           [0, 0, 4, 0, 0]])

    The anti-diagonal can be filled by reversing the order of elements
    using either `numpy.flipud` or `numpy.fliplr`.

    >>> a = mt.zeros((3, 3), int)
    >>> mt.fill_diagonal(mt.fliplr(a), [1,2,3])  # Horizontal flip
    >>> a.execute()
    array([[0, 0, 1],
           [0, 2, 0],
           [3, 0, 0]])
    >>> mt.fill_diagonal(mt.flipud(a), [1,2,3])  # Vertical flip
    >>> a.execute()
    array([[0, 0, 3],
           [0, 2, 0],
           [1, 0, 0]])

    Note that the order in which the diagonal is filled varies depending
    on the flip function.
    """

    if not isinstance(a, Tensor):
        raise TypeError(f"`a` should be a tensor, got {type(a)}")
    if a.ndim < 2:
        raise ValueError("array must be at least 2-d")
    if a.ndim > 2 and len(set(a.shape)) != 1:
        raise ValueError("All dimensions of input must be of equal length")

    # process val
    if isinstance(val, ENTITY_TYPE):
        val = astensor(val)
        if val.ndim > 1:
            val = val.ravel()
        val_input = val
    else:
        val = np.asarray(val)
        if val.ndim > 1:
            val = val.ravel()
        val_input = None

    op = TensorFillDiagonal(val=val, wrap=wrap, dtype=a.dtype)
    t = op(a, val=val_input)
    a.data = t.data



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Dict, Any


class ActorPoolConfig:
    __slots__ = ("_conf",)

    def __init__(self, conf: Dict = None):
        if conf is None:
            conf = dict()
        self._conf = conf
        if "pools" not in self._conf:
            self._conf["pools"] = dict()
        if "mapping" not in self._conf:
            self._conf["mapping"] = dict()
        if "metrics" not in self._conf:
            self._conf["metrics"] = dict()

    @property
    def n_pool(self):
        return len(self._conf["pools"])

    def add_pool_conf(
        self,
        process_index: int,
        label: str,
        internal_address: str,
        external_address: Union[str, List[str]],
        env: Dict = None,
        modules: List[str] = None,
        suspend_sigint: bool = False,
        use_uvloop: bool = False,
        logging_conf: Dict = None,
        kwargs: Dict = None,
    ):
        pools: Dict = self._conf["pools"]
        if not isinstance(external_address, list):
            external_address = [external_address]
        pools[process_index] = {
            "label": label,
            "internal_address": internal_address,
            "external_address": external_address,
            "env": env,
            "modules": modules,
            "suspend_sigint": suspend_sigint,
            "use_uvloop": use_uvloop,
            "logging_conf": logging_conf,
            "kwargs": kwargs or {},
        }

        mapping: Dict = self._conf["mapping"]
        for addr in external_address:
            mapping[addr] = internal_address

    def get_pool_config(self, process_index: int):
        return self._conf["pools"][process_index]

    def get_external_address(self, process_index: int) -> str:
        return self._conf["pools"][process_index]["external_address"][0]

    def get_process_indexes(self):
        return list(self._conf["pools"])

    def get_process_index(self, external_address: str):
        for process_index, conf in self._conf["pools"].items():
            if external_address in conf["external_address"]:
                return process_index
        raise ValueError(
            f"Cannot get process_index for {external_address}"
        )  # pragma: no cover

    def reset_pool_external_address(
        self,
        process_index: int,
        external_address: Union[str, List[str]],
    ):
        if not isinstance(external_address, list):
            external_address = [external_address]
        cur_pool_config = self._conf["pools"][process_index]
        internal_address = cur_pool_config["internal_address"]

        mapping: Dict = self._conf["mapping"]
        for addr in cur_pool_config["external_address"]:
            if internal_address == addr:
                # internal address may be the same as external address in Windows
                internal_address = external_address[0]
            mapping.pop(addr, None)

        cur_pool_config["external_address"] = external_address
        for addr in external_address:
            mapping[addr] = internal_address

    def get_external_addresses(self, label=None) -> List[str]:
        result = []
        for c in self._conf["pools"].values():
            if label is not None:
                if label == c["label"]:
                    result.append(c["external_address"][0])
            else:
                result.append(c["external_address"][0])
        return result

    @property
    def external_to_internal_address_map(self) -> Dict[str, str]:
        return self._conf["mapping"]

    def as_dict(self):
        return self._conf

    def add_metric_configs(self, metrics: Dict[str, Any]):
        if metrics:
            self._conf["metrics"].update(metrics)

    def get_metric_configs(self):
        return self._conf["metrics"]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import shutil
import subprocess
import sys
import warnings
from sysconfig import get_config_vars

from pkg_resources import parse_version
from setuptools import setup, Extension, Command

import numpy as np
from Cython.Build import cythonize
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.sdist import sdist

try:
    import distutils.ccompiler

    if sys.platform != "win32":
        from numpy.distutils.ccompiler import CCompiler_compile

        distutils.ccompiler.CCompiler.compile = CCompiler_compile
except ImportError:
    pass

# From https://github.com/pandas-dev/pandas/pull/24274:
# For mac, ensure extensions are built for macos 10.9 when compiling on a
# 10.9 system or above, overriding distuitls behaviour which is to target
# the version that python was built for. This may be overridden by setting
# MACOSX_DEPLOYMENT_TARGET before calling setup.py
if sys.platform == "darwin":
    if "MACOSX_DEPLOYMENT_TARGET" not in os.environ:
        current_system = platform.mac_ver()[0]
        python_target = get_config_vars().get(
            "MACOSX_DEPLOYMENT_TARGET", current_system
        )
        target_macos_version = "10.9"

        parsed_python_target = parse_version(python_target)
        parsed_current_system = parse_version(current_system)
        parsed_macos_version = parse_version(target_macos_version)
        if parsed_python_target <= parsed_macos_version <= parsed_current_system:
            os.environ["MACOSX_DEPLOYMENT_TARGET"] = target_macos_version


repo_root = os.path.dirname(os.path.abspath(__file__))


cythonize_kw = dict(language_level=sys.version_info[0])
cy_extension_kw = dict()
if os.environ.get("CYTHON_TRACE"):
    cy_extension_kw["define_macros"] = [
        ("CYTHON_TRACE_NOGIL", "1"),
        ("CYTHON_TRACE", "1"),
    ]
    cythonize_kw["compiler_directives"] = {"linetrace": True}

if "MSC" in sys.version:
    extra_compile_args = ["/std:c11", "/Ot", "/I" + os.path.join(repo_root, "misc")]
    cy_extension_kw["extra_compile_args"] = extra_compile_args
else:
    extra_compile_args = ["-O3"]
    if sys.platform != "darwin":
        # for macOS, we assume that C++ 11 is enabled by default
        extra_compile_args.append("-std=c++0x")
    cy_extension_kw["extra_compile_args"] = extra_compile_args


def _discover_pyx():
    exts = dict()
    for root, _, files in os.walk(os.path.join(repo_root, "mars")):
        for fn in files:
            if not fn.endswith(".pyx"):
                continue
            full_fn = os.path.relpath(os.path.join(root, fn), repo_root)
            mod_name = full_fn.replace(".pyx", "").replace(os.path.sep, ".")
            exts[mod_name] = Extension(mod_name, [full_fn], **cy_extension_kw)
    return exts


cy_extension_kw["include_dirs"] = [np.get_include()]
extensions_dict = _discover_pyx()
cy_extensions = list(extensions_dict.values())

extensions = cythonize(cy_extensions, **cythonize_kw) + [
    Extension(
        "mars.lib.mmh3",
        ["mars/lib/mmh3_src/mmh3module.cpp", "mars/lib/mmh3_src/MurmurHash3.cpp"],
    )
]


class ExtraCommandMixin:
    _extra_pre_commands = []

    def run(self):
        [self.run_command(cmd) for cmd in self._extra_pre_commands]
        super().run()

    @classmethod
    def register_pre_command(cls, cmd):
        cls._extra_pre_commands.append(cmd)


class CustomInstall(ExtraCommandMixin, install):
    pass


class CustomDevelop(ExtraCommandMixin, develop):
    pass


class CustomSDist(ExtraCommandMixin, sdist):
    pass


class BuildWeb(Command):
    """build_web command"""

    user_options = []
    _web_src_path = "mars/services/web/ui"
    _web_dest_path = "mars/services/web/static/bundle.js"
    _commands = [
        ["npm", "install"],
        ["npm", "run", "bundle"],
    ]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @classmethod
    def run(cls):
        if int(os.environ.get("NO_WEB_UI", "0")):
            return

        npm_path = shutil.which("npm")
        web_src_path = os.path.join(repo_root, *cls._web_src_path.split("/"))
        web_dest_path = os.path.join(repo_root, *cls._web_dest_path.split("/"))

        if not os.path.exists(web_src_path):
            return
        elif npm_path is None:
            if not os.path.exists(web_dest_path):
                warnings.warn("Cannot find NPM, may affect displaying Mars Web")
            return

        replacements = {"npm": npm_path}
        cmd_errored = False
        for cmd in cls._commands:
            cmd = [replacements.get(c, c) for c in cmd]
            proc_result = subprocess.run(cmd, cwd=web_src_path)
            if proc_result.returncode != 0:
                warnings.warn(f'Failed when running `{" ".join(cmd)}`')
                cmd_errored = True
                break
        if not cmd_errored:
            assert os.path.exists(cls._web_dest_path)


CustomInstall.register_pre_command("build_web")
CustomDevelop.register_pre_command("build_web")
CustomSDist.register_pre_command("build_web")


# Resolve path issue of versioneer
sys.path.append(repo_root)
versioneer = __import__("versioneer")


setup_options = dict(
    version=versioneer.get_version(),
    ext_modules=extensions,
    cmdclass=versioneer.get_cmdclass(
        {
            "build_web": BuildWeb,
            "install": CustomInstall,
            "develop": CustomDevelop,
            "sdist": CustomSDist,
        }
    ),
)
setup(**setup_options)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Iterable, List, Optional, Set, Tuple

from ...core import ChunkGraph, DAG, ChunkData
from ...resource import Resource
from ...serialization.serializables.field_type import TupleType
from ...serialization.serializables import (
    Serializable,
    StringField,
    ReferenceField,
    Int32Field,
    Int64Field,
    Float64Field,
    BoolField,
    AnyField,
    DictField,
    ListField,
    TupleField,
    FieldTypes,
)
from ...typing import BandType, ChunkType


class SubtaskStatus(Enum):
    pending = 0
    running = 1
    succeeded = 2
    errored = 3
    cancelled = 4

    @property
    def is_done(self) -> bool:
        return self in (
            SubtaskStatus.succeeded,
            SubtaskStatus.errored,
            SubtaskStatus.cancelled,
        )


class Subtask(Serializable):
    __slots__ = ("_repr", "_pure_depend_keys")

    subtask_id: str = StringField("subtask_id")
    subtask_name: str = StringField("subtask_name")
    session_id: str = StringField("session_id")
    task_id: str = StringField("task_id")
    chunk_graph: ChunkGraph = ReferenceField("chunk_graph", ChunkGraph)
    expect_bands: List[BandType] = ListField(
        "expect_bands", TupleType(FieldTypes.string, FieldTypes.string)
    )
    virtual: bool = BoolField("virtual")
    retryable: bool = BoolField("retryable")
    priority: Tuple[int, int] = TupleField("priority", FieldTypes.int32)
    rerun_time: int = Int32Field("rerun_time")
    extra_config: dict = DictField("extra_config")
    stage_id: str = StringField("stage_id")
    # chunks that need meta updated
    update_meta_chunks: List[ChunkType] = ListField(
        "update_meta_chunks", FieldTypes.reference(ChunkData)
    )
    # An unique and deterministic key for subtask compute logic. See logic_key in operator.py.
    logic_key: str = StringField("logic_key")
    # index for subtask with same compute logic.
    logic_index: int = Int32Field("logic_index")
    # parallelism for subtask with same compute logic.
    logic_parallelism: int = Int32Field("logic_parallelism")
    # subtask can only run in specified bands in `expect_bands`
    bands_specified: bool = BoolField("bands_specified")
    required_resource: Resource = AnyField("required_resource", Resource)

    def __init__(
        self,
        subtask_id: str = None,
        session_id: str = None,
        task_id: str = None,
        chunk_graph: ChunkGraph = None,
        subtask_name: str = None,
        expect_bands: List[BandType] = None,
        priority: Tuple[int, int] = None,
        virtual: bool = False,
        retryable: bool = True,
        rerun_time: int = 0,
        extra_config: dict = None,
        stage_id: str = None,
        update_meta_chunks: List[ChunkType] = None,
        logic_key: str = None,
        logic_index: int = None,
        logic_parallelism: int = None,
        bands_specified: bool = False,
        required_resource: Resource = None,
    ):
        super().__init__(
            subtask_id=subtask_id,
            subtask_name=subtask_name,
            session_id=session_id,
            task_id=task_id,
            chunk_graph=chunk_graph,
            expect_bands=expect_bands,
            priority=priority,
            virtual=virtual,
            retryable=retryable,
            rerun_time=rerun_time,
            extra_config=extra_config,
            stage_id=stage_id,
            update_meta_chunks=update_meta_chunks,
            logic_key=logic_key,
            logic_index=logic_index,
            logic_parallelism=logic_parallelism,
            bands_specified=bands_specified,
            required_resource=required_resource,
        )
        self._pure_depend_keys = None
        self._repr = None

    @property
    def expect_band(self):
        if self.expect_bands:
            return self.expect_bands[0]

    @property
    def pure_depend_keys(self) -> Set[str]:
        if self._pure_depend_keys is not None:
            return self._pure_depend_keys
        pure_dep_keys = set()
        for n in self.chunk_graph:
            pure_dep_keys.update(
                inp.key
                for inp, pure_dep in zip(n.inputs, n.op.pure_depends)
                if pure_dep
            )
        self._pure_depend_keys = pure_dep_keys
        return pure_dep_keys

    def __repr__(self):
        if self._repr is not None:
            return self._repr

        if self.chunk_graph:
            result_chunk_repr = " ".join(
                [
                    f"{type(chunk.op).__name__}({chunk.key})"
                    for chunk in self.chunk_graph.result_chunks
                ]
            )
        else:  # pragma: no cover
            result_chunk_repr = None
        self._repr = f"<Subtask id={self.subtask_id} results=[{result_chunk_repr}]>"
        return self._repr


class SubtaskResult(Serializable):
    subtask_id: str = StringField("subtask_id")
    session_id: str = StringField("session_id")
    task_id: str = StringField("task_id")
    stage_id: str = StringField("stage_id")
    status: SubtaskStatus = ReferenceField("status", SubtaskStatus)
    progress: float = Float64Field("progress", default=0.0)
    data_size: int = Int64Field("data_size", default=None)
    bands: List[BandType] = ListField("band", FieldTypes.tuple, default=None)
    error = AnyField("error", default=None)
    traceback = AnyField("traceback", default=None)
    # The following is the execution information of the subtask
    execution_start_time: float = Float64Field("execution_start_time")
    execution_end_time: float = Float64Field("execution_end_time")

    def update(self, result: Optional["SubtaskResult"]):
        if result and result.bands:
            bands = self.bands or []
            self.bands = sorted(set(bands + result.bands))
            self.execution_start_time = result.execution_start_time
            if hasattr(result, "execution_end_time"):
                self.execution_end_time = result.execution_end_time
        return self


class SubtaskGraph(DAG, Iterable[Subtask]):
    """
    Subtask graph.
    """

    @classmethod
    def _extract_operands(cls, node: Subtask):
        from ...core.operand import Fetch, FetchShuffle

        for node in node.chunk_graph:
            if isinstance(node.op, (Fetch, FetchShuffle)):
                continue
            yield node.op



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import recursive_tile
from ...serialization.serializables import AnyField, BoolField, Int64Field
from ...tensor.core import TENSOR_TYPE, TENSOR_CHUNK_TYPE
from ..core import SERIES_TYPE, SERIES_CHUNK_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_empty_df, parse_index


class DataFrameInsert(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.INSERT

    _loc = Int64Field("loc")
    _column = AnyField("column")
    _value = AnyField("value")
    _allow_duplicates = BoolField("allow_duplicates")

    def __init__(self, loc=None, column=None, value=None, allow_duplicates=None, **kw):
        super().__init__(
            _loc=loc,
            _column=column,
            _value=value,
            _allow_duplicates=allow_duplicates,
            **kw,
        )

    @property
    def loc(self) -> int:
        return self._loc

    @property
    def column(self):
        return self._column

    @property
    def value(self):
        return self._value

    @property
    def allow_duplicates(self):
        return self._allow_duplicates

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if len(inputs) > 1:
            self._value = self._inputs[-1]

    def __call__(self, df):
        inputs = [df]
        if isinstance(self.value, (SERIES_TYPE, TENSOR_TYPE)):
            value_dtype = self.value.dtype
            inputs.append(self.value)
        else:
            value_dtype = pd.Series(self.value).dtype

        empty_df = build_empty_df(df.dtypes)
        empty_df.insert(
            loc=self.loc,
            column=self.column,
            allow_duplicates=self.allow_duplicates,
            value=pd.Series([], dtype=value_dtype),
        )

        params = df.params
        params["columns_value"] = parse_index(empty_df.columns, store_data=True)
        params["dtypes"] = empty_df.dtypes
        params["shape"] = (df.shape[0], df.shape[1] + 1)
        return self.new_dataframe(inputs, **params)

    @classmethod
    def tile(cls, op: "DataFrameInsert"):
        inp = op.inputs[0]
        value = op.value
        if isinstance(value, (SERIES_TYPE, TENSOR_TYPE)):
            value = yield from recursive_tile(value.rechunk({0: inp.nsplits[0]}))
        out = op.outputs[0]

        chunk_bounds = np.cumsum((0,) + inp.nsplits[1])
        chunk_bounds[-1] += 1

        chunks = []
        new_split = list(inp.nsplits[1])
        chunk_dtypes = None
        chunk_columns_value = None
        for c in inp.chunks:
            left_bound = int(chunk_bounds[c.index[1]])
            right_bound = int(chunk_bounds[c.index[1] + 1])
            if left_bound > op.loc or right_bound <= op.loc:
                chunks.append(c)
                continue

            if chunk_dtypes is None:
                new_split[c.index[1]] = inp.nsplits[1][c.index[1]] + 1

                if isinstance(value, (SERIES_TYPE, TENSOR_TYPE)):
                    value_dtype = value.dtype
                else:
                    value_dtype = pd.Series(value).dtype

                empty_df = build_empty_df(c.dtypes)
                empty_df.insert(
                    loc=op.loc - left_bound,
                    column=op.column,
                    allow_duplicates=op.allow_duplicates,
                    value=pd.Series([], dtype=value_dtype),
                )

                chunk_dtypes = empty_df.dtypes
                chunk_columns_value = parse_index(chunk_dtypes.index, store_data=True)

            params = c.params
            params["columns_value"] = chunk_columns_value
            params["dtypes"] = chunk_dtypes
            params["shape"] = (c.shape[0], c.shape[1] + 1)

            new_op = op.copy().reset_key()
            new_op._loc = op.loc - left_bound

            if isinstance(value, (SERIES_TYPE, TENSOR_TYPE)):
                inputs = [c, value.chunks[c.index[0]]]
            else:
                inputs = [c]
            chunks.append(new_op.new_chunk(inputs, **params))

        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            [inp],
            chunks=chunks,
            nsplits=(inp.nsplits[0], tuple(new_split)),
            **out.params,
        )

    @classmethod
    def execute(cls, ctx, op: "DataFrameInsert"):
        input_ = ctx[op.inputs[0].key]
        value = op.value
        if isinstance(value, (SERIES_CHUNK_TYPE, TENSOR_CHUNK_TYPE)):
            value = ctx[value.key]
        ctx[op.outputs[0].key] = copied = input_.copy()
        copied.insert(
            loc=op.loc,
            column=op.column,
            allow_duplicates=op.allow_duplicates,
            value=value,
        )


def df_insert(df, loc, column, value, allow_duplicates=False):
    """
    Insert column into DataFrame at specified location.

    Raises a ValueError if `column` is already contained in the DataFrame,
    unless `allow_duplicates` is set to True.

    Parameters
    ----------
    loc : int
        Insertion index. Must verify 0 <= loc <= len(columns).
    column : str, number, or hashable object
        Label of the inserted column.
    value : int, Series, or array-like
    allow_duplicates : bool, optional
    """
    if isinstance(value, TENSOR_TYPE) and value.ndim > 1:
        raise ValueError(
            f"Wrong number of items passed {value.ndim}, placement implies 1"
        )

    op = DataFrameInsert(
        loc=loc, column=column, value=value, allow_duplicates=allow_duplicates
    )
    out_df = op(df)
    df.data = out_df.data



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core.operand import OperandStage
from .topk import _validate_topk_arguments, TensorTopk


def argtopk(
    a,
    k,
    axis=-1,
    largest=True,
    sorted=True,
    order=None,
    parallel_kind="auto",
    psrs_kinds=None,
):
    (
        a,
        k,
        axis,
        largest,
        sorted,
        order,
        parallel_kind,
        psrs_kinds,
    ) = _validate_topk_arguments(
        a, k, axis, largest, sorted, order, parallel_kind, psrs_kinds
    )
    op = TensorTopk(
        k=k,
        axis=axis,
        largest=largest,
        sorted=sorted,
        parallel_kind=parallel_kind,
        psrs_kinds=psrs_kinds,
        dtype=a.dtype,
        return_value=False,
        return_indices=True,
        stage=OperandStage.agg,
    )
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialization.serializables import BoolField, ListField, Int32Field
from ...tensor.core import TENSOR_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index


class DataFrameFromRecords(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_FROM_RECORDS

    columns = ListField("columns", default=None)
    exclude = ListField("exclude", default=None)
    coerce_float = BoolField("coerce_float", default=False)
    nrows = Int32Field("nrows", default=None)

    def __init__(self, index=None, columns=None, **kw):
        if index is not None or columns is not None:
            raise NotImplementedError("Specifying index value is not supported for now")
        super().__init__(columns=columns, _output_types=[OutputType.dataframe], **kw)

    def __call__(self, data):
        if self.nrows is None:
            nrows = data.shape[0]
        else:
            nrows = self.nrows
        index_value = parse_index(pd.RangeIndex(start=0, stop=nrows))
        dtypes = pd.Series(dict((k, np.dtype(v)) for k, v in data.dtype.descr))
        columns_value = parse_index(pd.Index(data.dtype.names), store_data=True)
        return self.new_dataframe(
            [data],
            (data.shape[0], len(data.dtype.names)),
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )

    @classmethod
    def tile(cls, op):
        df = op.outputs[0]
        tensor = op.inputs[0]

        nsplit_acc = np.cumsum(tensor.nsplits[0])
        out_chunks = []
        for chunk in tensor.chunks:
            begin_index = nsplit_acc[chunk.index[0]] - chunk.shape[0]
            end_index = nsplit_acc[chunk.index[0]]
            chunk_index_value = parse_index(
                pd.RangeIndex(start=begin_index, stop=end_index)
            )

            # Here the `new_chunk` is tricky:
            #
            # We can construct tensor that have identifcal chunks, for example, from `mt.ones(...)`, we know
            # that after tiling the chunk of the same shape (but at different position) in `mt.ones` is indeed
            # the same chunk (has the same key)!
            #
            # Thus, when we construct dataframe from such tensor, we will have dataframe chunks that only differ
            # in `index_value`. However the `index_value` field won't be used to calculate the chunk key of
            # the dataframe chunk, thus `new_chunk` generated the same keys for those indeed different chunks
            # (they have different `index_values`).
            #
            # Here, we construct new chunk with some unique `_extra_params` to make the `new_chunk` work as
            # expected.
            chunk_op = op.copy().reset_key()
            chunk_op.extra_params["begin_index"] = begin_index
            chunk_op.extra_params["end_index"] = end_index
            out_chunk = chunk_op.new_chunk(
                [chunk],
                shape=(chunk.shape[0], df.shape[1]),
                index=(chunk.index[0], 0),
                dtypes=df.dtypes,
                index_value=chunk_index_value,
                columns_value=df.columns_value,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_dataframes(
            [tensor],
            df.shape,
            dtypes=df.dtypes,
            index_value=df.index_value,
            columns_value=df.columns_value,
            chunks=out_chunks,
            nsplits=[tensor.nsplits[0], [df.shape[1]]],
        )

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        ctx[chunk.key] = pd.DataFrame.from_records(
            ctx[op.inputs[0].key],
            index=chunk.index_value.to_pandas(),
            columns=chunk.columns_value.to_pandas(),
            exclude=op.exclude,
            coerce_float=op.coerce_float,
            nrows=op.nrows,
        )


def from_records(
    data,
    index=None,
    exclude=None,
    columns=None,
    coerce_float=False,
    nrows=None,
    gpu=None,
    sparse=False,
    **kw
):
    if isinstance(data, np.ndarray):
        from .dataframe import from_pandas

        return from_pandas(
            pd.DataFrame.from_records(
                data,
                index=index,
                exclude=exclude,
                columns=columns,
                coerce_float=coerce_float,
                nrows=nrows,
            ),
            **kw
        )
    elif isinstance(data, TENSOR_TYPE):
        if data.dtype.names is None:
            raise TypeError("Not a tensor with structured dtype {0}", data.dtype)
        if data.ndim != 1:
            raise ValueError(
                "Not a tensor with non 1-D structured dtype {0}", data.shape
            )

        op = DataFrameFromRecords(
            index=None,
            exclude=exclude,
            columns=columns,
            coerce_float=coerce_float,
            nrows=nrows,
            gpu=gpu,
            sparse=sparse,
            **kw
        )
        return op(data)
    else:
        raise TypeError("Not support create DataFrame from {0}", type(data))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialization.serializables import StringField
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameStrConcat(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = OperandDef.STR_CONCAT
    _func_name = "str_concat"

    _sep = StringField("sep")
    _na_rep = StringField("na_rep")

    def __init__(self, sep=None, na_rep=None, **kw):
        super().__init__(_sep=sep, _na_rep=na_rep, **kw)

    @property
    def sep(self):
        return self._sep

    @property
    def na_rep(self):
        return self._na_rep

    def get_reduction_args(self, axis=None):
        return dict(sep=self._sep, na_rep=self._na_rep)

    @property
    def is_atomic(self):
        return True

    @classmethod
    def get_reduction_callable(cls, op):
        sep, na_rep = op.sep, op.na_rep

        def str_concat(obj):
            return build_str_concat_object(obj, sep=sep, na_rep=na_rep)

        return str_concat


def build_str_concat_object(df, sep=None, na_rep=None):
    output_type = OutputType.series if df.ndim == 2 else OutputType.scalar
    op = DataFrameStrConcat(sep=sep, na_rep=na_rep, output_types=[output_type])
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType, recursive_tile
from ...serialization.serializables import BoolField
from ..datasource.dataframe import from_pandas
from ..indexing.iloc import DataFrameIlocGetItem, SeriesIlocGetItem
from ..utils import parse_index, standardize_range_index
from ..operands import (
    DataFrameOperand,
    DataFrameOperandMixin,
    DATAFRAME_TYPE,
    SERIES_TYPE,
)


class DataFrameAppend(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.APPEND

    _ignore_index = BoolField("ignore_index")
    _verify_integrity = BoolField("verify_integrity")
    _sort = BoolField("sort")

    def __init__(
        self,
        ignore_index=None,
        verify_integrity=None,
        sort=None,
        output_types=None,
        **kw,
    ):
        super().__init__(
            _ignore_index=ignore_index,
            _verify_integrity=verify_integrity,
            _sort=sort,
            _output_types=output_types,
            **kw,
        )

    @property
    def ignore_index(self):
        return self._ignore_index

    @property
    def verify_integrity(self):
        return self._verify_integrity

    @property
    def sort(self):
        return self._sort

    @classmethod
    def _tile_dataframe(cls, op):
        out_df = op.outputs[0]
        inputs = op.inputs
        first_df, others = inputs[0], inputs[1:]
        column_splits = first_df.nsplits[1]
        new_others = []
        for item in others:
            r = yield from recursive_tile(item.rechunk({1: column_splits}))
            new_others.append(r)
        others = new_others
        out_chunks = []
        nsplits = [[], list(first_df.nsplits[1])]
        row_index = 0
        for df in [first_df] + others:
            for c in df.chunks:
                index = (c.index[0] + row_index, c.index[1])
                iloc_op = DataFrameIlocGetItem(indexes=[slice(None)] * 2)
                out_chunks.append(
                    iloc_op.new_chunk(
                        [c],
                        shape=c.shape,
                        index=index,
                        dtypes=c.dtypes,
                        index_value=c.index_value,
                        columns_value=c.columns_value,
                    )
                )
            nsplits[0] += df.nsplits[0]
            row_index += len(df.nsplits[0])
        if op.ignore_index:
            yield out_chunks
            out_chunks = standardize_range_index(out_chunks)

        nsplits = tuple(tuple(n) for n in nsplits)
        new_op = op.copy()
        return new_op.new_dataframes(
            op.inputs,
            out_df.shape,
            nsplits=nsplits,
            chunks=out_chunks,
            dtypes=out_df.dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
        )

    @classmethod
    def _tile_series(cls, op):
        out_series = op.outputs[0]
        inputs = op.inputs
        first_series, others = inputs[0], inputs[1:]
        out_chunks = []
        nsplits = ()
        row_index = 0
        for series in [first_series] + others:
            for c in series.chunks:
                index = (c.index[0] + row_index,)
                iloc_op = SeriesIlocGetItem(indexes=(slice(None),))
                out_chunks.append(
                    iloc_op.new_chunk(
                        [c],
                        shape=c.shape,
                        index=index,
                        index_value=c.index_value,
                        dtype=c.dtype,
                        name=c.name,
                    )
                )
            nsplits += series.nsplits[0]
            row_index += len(series.nsplits[0])

        if op.ignore_index:
            yield out_chunks
            out_chunks = standardize_range_index(out_chunks)

        nsplits = (tuple(nsplits),)
        new_op = op.copy()
        return new_op.new_seriess(
            op.inputs,
            out_series.shape,
            nsplits=nsplits,
            chunks=out_chunks,
            dtype=out_series.dtype,
            index_value=out_series.index_value,
            name=out_series.name,
        )

    @classmethod
    def tile(cls, op):
        if op.output_types[0] == OutputType.dataframe:
            return (yield from cls._tile_dataframe(op))
        else:
            return (yield from cls._tile_series(op))

    def _call_dataframe(self, df, other):
        if isinstance(other, DATAFRAME_TYPE):
            shape = (df.shape[0] + other.shape[0], df.shape[1])
            inputs = [df, other]
            if self.ignore_index:
                index_value = parse_index(pd.RangeIndex(shape[0]))
            else:
                index_value = parse_index(
                    df.index_value.to_pandas().append(other.index_value.to_pandas())
                )
        elif isinstance(other, list):
            row_length = df.shape[0]
            index = df.index_value.to_pandas()
            for item in other:
                if not isinstance(item, DATAFRAME_TYPE):  # pragma: no cover
                    raise ValueError(f"Invalid type {type(item)} to append")
                row_length += item.shape[0]
                index = index.append(item.index_value.to_pandas())
            shape = (row_length, df.shape[1])
            if self.ignore_index:  # pragma: no cover
                index_value = parse_index(pd.RangeIndex(shape[0]))
            else:
                index_value = parse_index(index)
            inputs = [df] + other
        else:  # pragma: no cover
            raise ValueError(f"Invalid type {type(other)} to append")
        return self.new_dataframe(
            inputs,
            shape=shape,
            dtypes=df.dtypes,
            index_value=index_value,
            columns_value=df.columns_value,
        )

    def _call_series(self, df, other):
        if isinstance(other, SERIES_TYPE):
            shape = (df.shape[0] + other.shape[0],)
            inputs = [df, other]
            if self.ignore_index:
                index_value = parse_index(pd.RangeIndex(shape[0]))
            else:
                index_value = parse_index(
                    df.index_value.to_pandas().append(other.index_value.to_pandas())
                )
        elif isinstance(other, list):
            row_length = df.shape[0]
            index = df.index_value.to_pandas()
            for item in other:
                if not isinstance(item, SERIES_TYPE):  # pragma: no cover
                    raise ValueError(f"Invalid type {type(item)} to append")
                row_length += item.shape[0]
                index = index.append(item.index_value.to_pandas())
            shape = (row_length,)
            if self.ignore_index:  # pragma: no cover
                index_value = parse_index(pd.RangeIndex(shape[0]))
            else:
                index_value = parse_index(index)
            inputs = [df] + other
        else:  # pragma: no cover
            raise ValueError(f"Invalid type {type(other)} to append")
        return self.new_series(
            inputs, shape=shape, dtype=df.dtype, index_value=index_value, name=df.name
        )

    @classmethod
    def execute(cls, ctx, op):
        first, others = ctx[op.inputs[0].key], [ctx[inp.key] for inp in op.inputs[1:]]
        r = first.append(others, verify_integrity=op.verify_integrity, sort=op.sort)
        ctx[op.outputs[0].key] = r

    def __call__(self, df, other):
        if isinstance(df, DATAFRAME_TYPE):
            self.output_types = [OutputType.dataframe]
            return self._call_dataframe(df, other)
        else:
            self.output_types = [OutputType.series]
            return self._call_series(df, other)


def append(df, other, ignore_index=False, verify_integrity=False, sort=False):
    if verify_integrity or sort:  # pragma: no cover
        raise NotImplementedError("verify_integrity and sort are not supported now")
    if isinstance(other, dict):
        other = from_pandas(pd.DataFrame(dict((k, [v]) for k, v in other.items())))
    op = DataFrameAppend(
        ignore_index=ignore_index, verify_integrity=verify_integrity, sort=sort
    )
    return op(df, other)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any, BinaryIO, Dict, List, Optional, TextIO, Union

import numpy as np

from .... import opcodes as OperandDef
from ....core.context import get_context
from ....remote.run_script import RunScript, _extract_inputs
from ....serialization.serializables import Int32Field, StringField
from ....typing import SessionType, TileableType
from ....utils import to_binary
from ..utils import pick_workers


class RunPyTorch(RunScript):
    _op_type_ = OperandDef.RUN_PYTORCH

    # used for chunk op
    _master_port = Int32Field("master_port")
    _master_addr = StringField("master_addr")
    _rank = Int32Field("rank")
    _init_method = StringField("init_method")

    def __init__(
        self, master_port=None, master_addr=None, init_method=None, gpu=None, **kw
    ):
        super().__init__(
            _master_port=master_port,
            _master_addr=master_addr,
            _init_method=init_method,
            gpu=gpu,
            **kw
        )

    @property
    def master_port(self):
        return self._master_port

    @property
    def master_addr(self):
        return self._master_addr

    @property
    def init_method(self):
        return self._init_method

    @classmethod
    def tile(cls, op):
        ctx = get_context()

        workers = pick_workers(ctx.get_worker_addresses(), op.world_size)
        data, input_chunks = cls._get_chunk_data(op)

        out_chunks = []
        for i in range(op.world_size):
            chunk_op = op.copy().reset_key()
            chunk_op._data = data
            chunk_op.expect_worker = workers[i]
            if op.init_method is None:
                chunk_op._master_port = op.master_port
                chunk_op._master_addr = workers[0].split(":", 1)[0]
            chunk_op._rank = i
            chunk_op._init_method = op.init_method
            out_chunks.append(chunk_op.new_chunk(input_chunks, index=(i,)))

        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs,
            chunks=out_chunks,
            nsplits=(tuple(np.nan for _ in range(len(out_chunks))),),
        )

    @classmethod
    def _build_envs(cls, ctx, op):
        envs = super()._build_envs(ctx, op)
        if op.master_port is not None:
            envs["MASTER_PORT"] = str(op.master_port)
        if op.master_addr is not None:
            envs["MASTER_ADDR"] = str(op.master_addr)
        return envs

    @classmethod
    def execute(cls, ctx, op):
        assert ctx.local_address.split(":")[0] == op.expect_worker.split(":")[0]

        super().execute(ctx, op)


def run_pytorch_script(
    script: Union[bytes, str, BinaryIO, TextIO],
    n_workers: int,
    data: Dict[str, TileableType] = None,
    gpu: Optional[bool] = None,
    command_argv: List[str] = None,
    retry_when_fail: bool = False,
    session: SessionType = None,
    run_kwargs: Dict[str, Any] = None,
    port: int = None,
):
    """
    Run PyTorch script in Mars cluster.

    Parameters
    ----------
    script: str or file-like object
        Script to run
    n_workers : int
        Number of PyTorch workers
    data : dict
        Variable name to data.
    gpu : bool
        Run PyTorch script on GPU
    command_argv : list
        Extra command args for script
    retry_when_fail : bool
        If True, retry when function failed.
    session
        Mars session, if not provided, will use default one.
    run_kwargs : dict
        Extra kwargs for `session.run`.
    port : int
        Port of PyTorch worker or ps, will automatically increase for the same worker

    Returns
    -------
    status
        return {'status': 'ok'} if succeeded, or error raised
    """
    if int(n_workers) <= 0:
        raise ValueError("n_workers should be at least 1")
    if hasattr(script, "read"):
        code = script.read()
    else:
        with open(os.path.abspath(script), "rb") as f:
            code = f.read()

    inputs = _extract_inputs(data)
    port = 29500 if port is None else port
    op = RunPyTorch(
        data=data,
        code=to_binary(code),
        world_size=int(n_workers),
        retry_when_fail=retry_when_fail,
        gpu=gpu,
        master_port=port,
        command_args=command_argv,
    )
    return op(inputs).execute(session=session, **(run_kwargs or {}))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List, Tuple, Dict, Type

from ...core import ChunkGraph, ChunkType, OperandType
from ...utils import build_fuse_chunk


class RuntimeOptimizer(ABC):
    engine = None

    def __init__(self, graph: ChunkGraph):
        self._graph = graph

    @classmethod
    @abstractmethod
    def is_available(cls) -> bool:
        """
        Check this optimizer is available.

        Returns
        -------
        is_available : bool
            Available.
        """

    @abstractmethod
    def optimize(self):
        """
        Optimize chunk graph.
        """

    def _fuse_nodes(
        self, fuses: List[List[ChunkType]], fuse_cls: OperandType
    ) -> Tuple[List[List[ChunkType]], List[ChunkType]]:
        graph = self._graph
        fused_nodes = []

        for fuse in fuses:
            head_node = fuse[0]
            tail_node = fuse[-1]

            fused_chunk = build_fuse_chunk(
                fuse, fuse_cls, op_kw={"dtype": tail_node.dtype}
            ).data
            graph.add_node(fused_chunk)
            for node in graph.iter_successors(tail_node):
                graph.add_edge(fused_chunk, node)
            for node in graph.iter_predecessors(head_node):
                graph.add_edge(node, fused_chunk)
            for node in fuse:
                graph.remove_node(node)
            fused_nodes.append(fused_chunk)

            try:
                # check tail node if it's in results
                i = graph.results.index(tail_node)
                graph.results[i] = fused_chunk
            except ValueError:
                pass

        return fuses, fused_nodes


_engine_to_optimizers: Dict[str, Type[RuntimeOptimizer]] = dict()


def register_optimizer(optimizer_cls: Type[RuntimeOptimizer]):
    _engine_to_optimizers[optimizer_cls.engine] = optimizer_cls
    return optimizer_cls


def optimize(graph: ChunkGraph, engines: List[str] = None) -> ChunkGraph:
    if engines is None:
        engines = ["numexpr", "cupy"]

    for engine in engines:
        optimizer_cls = _engine_to_optimizers[engine]
        optimizer = optimizer_cls(graph)
        if not optimizer.is_available():
            continue
        optimizer.optimize()

    return graph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..base import atleast_3d
from .concatenate import concatenate


def dstack(tup):
    """
    Stack tensors in sequence depth wise (along third axis).

    This is equivalent to concatenation along the third axis after 2-D tensors
    of shape `(M,N)` have been reshaped to `(M,N,1)` and 1-D arrays of shape
    `(N,)` have been reshaped to `(1,N,1)`. Rebuilds arrays divided by
    `dsplit`.

    This function makes most sense for arrays with up to 3 dimensions. For
    instance, for pixel-data with a height (first axis), width (second axis),
    and r/g/b channels (third axis). The functions `concatenate`, `stack` and
    `block` provide more general stacking and concatenation operations.

    Parameters
    ----------
    tup : sequence of tensors
        The tensors must have the same shape along all but the third axis.
        1-D or 2-D arrays must have the same shape.

    Returns
    -------
    stacked : Tensor
        The array formed by stacking the given tensors, will be at least 3-D.

    See Also
    --------
    stack : Join a sequence of tensors along a new axis.
    vstack : Stack along first axis.
    hstack : Stack along second axis.
    concatenate : Join a sequence of arrays along an existing axis.
    dsplit : Split tensor along third axis.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array((1,2,3))
    >>> b = mt.array((2,3,4))
    >>> mt.dstack((a,b)).execute()
    array([[[1, 2],
            [2, 3],
            [3, 4]]])

    >>> a = mt.array([[1],[2],[3]])
    >>> b = mt.array([[2],[3],[4]])
    >>> mt.dstack((a,b)).execute()
    array([[[1, 2]],
           [[2, 3]],
           [[3, 4]]])

    """
    return concatenate([atleast_3d(t) for t in tup], axis=2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core import ExecutableTuple
from ..utils import broadcast_shape
from ..datasource import tensor as astensor
from .broadcast_to import broadcast_to


def broadcast_arrays(*args, **kwargs):
    """
    Broadcast any number of arrays against each other.

    Parameters
    ----------
    `*args` : array_likes
        The tensors to broadcast.

    Returns
    -------
    broadcasted : list of tensors

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([[1,2,3]])
    >>> y = mt.array([[1],[2],[3]])
    >>> mt.broadcast_arrays(x, y).execute()
    [array([[1, 2, 3],
           [1, 2, 3],
           [1, 2, 3]]), array([[1, 1, 1],
           [2, 2, 2],
           [3, 3, 3]])]

    """
    if kwargs:
        raise TypeError(
            "broadcast_arrays() got an unexpected keyword "
            f"argument {next(iter(kwargs.keys()))!r}"
        )

    args = [astensor(arg) for arg in args]

    shape = broadcast_shape(*[arg.shape for arg in args])
    return ExecutableTuple([broadcast_to(a, shape) for a in args])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, recursive_tile
from ...serialization.serializables import Int32Field, TupleField, AnyField, KeyField
from ...utils import has_unknown_shape
from ..datasource import tensor as astensor
from ..operands import TensorHasInput, TensorOperandMixin
from ..utils import filter_inputs, validate_axis, calc_object_length


class TensorInsert(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.INSERT

    _index_obj = AnyField("index_obj")
    _values = AnyField("values")
    _axis = Int32Field("axis")
    _input = KeyField("input")

    # for chunk
    _range_on_axis = TupleField("range_on_axis")

    def __init__(
        self, index_obj=None, values=None, axis=None, range_on_axis=None, **kw
    ):
        super().__init__(
            _index_obj=index_obj,
            _values=values,
            _axis=axis,
            _range_on_axis=range_on_axis,
            **kw
        )

    @property
    def index_obj(self):
        return self._index_obj

    @property
    def values(self):
        return self._values

    @property
    def axis(self):
        return self._axis

    @property
    def range_on_axis(self):
        return self._range_on_axis

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs[1:])
        if isinstance(self._index_obj, ENTITY_TYPE):
            self._index_obj = next(inputs_iter)
        if isinstance(self._values, ENTITY_TYPE):
            self._values = next(inputs_iter)

    @classmethod
    def tile(cls, op: "TensorInsert"):
        inp = op.inputs[0]
        axis = op.axis
        if axis is None:
            inp = yield from recursive_tile(inp.flatten())
            axis = 0
        else:
            new_splits = [s if i == axis else sum(s) for i, s in enumerate(inp.nsplits)]
            inp = yield from recursive_tile(inp.rechunk(new_splits))

        if has_unknown_shape(inp):
            yield

        index_obj = op.index_obj
        values = op.values
        if isinstance(values, ENTITY_TYPE):
            # if values is Mars type, we rechunk it into one chunk and
            # all insert chunks depend on it
            values = yield from recursive_tile(values.rechunk(values.shape))

        nsplits_on_axis = []
        if isinstance(index_obj, int):
            splits = inp.nsplits[axis]
            cum_splits = np.cumsum([0] + list(splits))
            # add 1 for last split
            cum_splits[-1] = cum_splits[-1] + 1
            in_idx = cum_splits.searchsorted(index_obj, side="right") - 1
            out_chunks = []
            for chunk in inp.chunks:
                if chunk.index[axis] == in_idx:
                    chunk_op = op.copy().reset_key()
                    chunk_op._index_obj = index_obj - cum_splits[in_idx]
                    if isinstance(values, ENTITY_TYPE):
                        chunk_values = values.chunks[0]
                    else:
                        chunk_values = values
                    inputs = filter_inputs([chunk, chunk_values])
                    shape = tuple(
                        s + calc_object_length(index_obj) if i == axis else s
                        for i, s in enumerate(chunk.shape)
                    )
                    out_chunks.append(
                        chunk_op.new_chunk(inputs, shape=shape, index=chunk.index)
                    )
                    nsplits_on_axis.append(shape[axis])
                else:
                    out_chunks.append(chunk)
                    nsplits_on_axis.append(chunk.shape[axis])
        elif isinstance(index_obj, ENTITY_TYPE):
            index_obj = yield from recursive_tile(index_obj.rechunk(index_obj.shape))
            offset = 0
            out_chunks = []
            for chunk in inp.chunks:
                chunk_op = op.copy().reset_key()
                chunk_op._index_obj = index_obj.chunks[0]
                if isinstance(values, ENTITY_TYPE):
                    chunk_values = values.chunks[0]
                else:
                    chunk_values = values
                chunk_op._values = chunk_values
                if chunk.index[axis] + 1 == len(inp.nsplits[axis]):
                    # the last chunk on axis
                    chunk_op._range_on_axis = (offset, offset + chunk.shape[axis] + 1)
                else:
                    chunk_op._range_on_axis = (offset, offset + chunk.shape[axis])
                shape = tuple(
                    np.nan if j == axis else s for j, s in enumerate(chunk.shape)
                )
                inputs = filter_inputs([chunk, index_obj.chunks[0], chunk_values])
                out_chunks.append(
                    chunk_op.new_chunk(inputs, shape=shape, index=chunk.index)
                )
                offset += chunk.shape[axis]
                nsplits_on_axis.append(np.nan)
        else:
            # index object is slice or sequence of ints
            if isinstance(index_obj, slice):
                index_obj = range(
                    index_obj.start or 0, index_obj.stop, index_obj.step or 1
                )
            splits = inp.nsplits[axis]
            cum_splits = np.cumsum([0] + list(splits))
            # add 1 for last split
            cum_splits[-1] = cum_splits[-1] + 1
            chunk_idx_params = [[[], []] for _ in splits]
            for i, int_idx in enumerate(index_obj):
                in_idx = cum_splits.searchsorted(int_idx, side="right") - 1
                chunk_idx_params[in_idx][0].append(int_idx - cum_splits[in_idx])
                chunk_idx_params[in_idx][1].append(i)

            out_chunks = []
            offset = 0
            for chunk in inp.chunks:
                idx_on_axis = chunk.index[axis]
                if len(chunk_idx_params[idx_on_axis][0]) > 0:
                    chunk_op = op.copy().reset_key()
                    chunk_index_obj = chunk_idx_params[idx_on_axis][0]
                    shape = tuple(
                        s + len(chunk_index_obj) if j == axis else s
                        for j, s in enumerate(chunk.shape)
                    )
                    if isinstance(values, int):
                        chunk_op._index_obj = chunk_index_obj
                        out_chunks.append(
                            chunk_op.new_chunk([chunk], shape=shape, index=chunk.index)
                        )
                    elif isinstance(values, ENTITY_TYPE):
                        chunk_op._values = values.chunks[0]
                        if chunk.index[axis] + 1 == len(inp.nsplits[axis]):
                            chunk_op._range_on_axis = (
                                offset,
                                offset + chunk.shape[axis] + 1,
                            )
                        else:
                            chunk_op._range_on_axis = (
                                offset,
                                offset + chunk.shape[axis],
                            )
                        out_chunks.append(
                            chunk_op.new_chunk(
                                [chunk, values.chunks[0]],
                                shape=shape,
                                index=chunk.index,
                            )
                        )
                        offset += chunk.shape[axis]
                    else:
                        chunk_op._index_obj = chunk_index_obj
                        values = np.asarray(values)
                        to_shape = [
                            calc_object_length(index_obj, chunk.shape[axis])
                        ] + [s for j, s in enumerate(inp.shape) if j != axis]
                        if all(j == k for j, k in zip(to_shape, values.shape)):
                            chunk_values = np.asarray(values)[
                                chunk_idx_params[idx_on_axis][1]
                            ]
                            chunk_op._values = chunk_values
                            out_chunks.append(
                                chunk_op.new_chunk(
                                    [chunk], shape=shape, index=chunk.index
                                )
                            )
                        else:
                            out_chunks.append(
                                chunk_op.new_chunk(
                                    [chunk], shape=shape, index=chunk.index
                                )
                            )

                    nsplits_on_axis.append(shape[axis])
                else:
                    out_chunks.append(chunk)
                    nsplits_on_axis.append(chunk.shape[axis])

        nsplits = tuple(
            s if i != axis else tuple(nsplits_on_axis)
            for i, s in enumerate(inp.nsplits)
        )
        out = op.outputs[0]
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            shape=out.shape,
            order=out.order,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def execute(cls, ctx, op: "TensorInsert"):
        inp = ctx[op.input.key]
        index_obj = (
            ctx[op.index_obj.key] if hasattr(op.index_obj, "key") else op.index_obj
        )
        values = ctx[op.values.key] if hasattr(op.values, "key") else op.values
        if op.range_on_axis is None:
            ctx[op.outputs[0].key] = np.insert(inp, index_obj, values, axis=op.axis)
        else:
            if isinstance(index_obj, slice):
                index_obj = np.arange(
                    index_obj.step or 0, index_obj.stop, index_obj.step or 1
                )
            else:
                index_obj = np.array(index_obj)
            values = np.asarray(values)

            part_index = [
                i
                for i, idx in enumerate(index_obj)
                if ((idx >= op.range_on_axis[0]) and idx < op.range_on_axis[1])
            ]
            if (
                (values.ndim > 0)
                and len(index_obj) == len(values)
                and (values[0].ndim > 0 or inp.ndim == 1)
            ):
                ctx[op.outputs[0].key] = np.insert(
                    inp,
                    index_obj[part_index] - op.range_on_axis[0],
                    values[part_index],
                    axis=op.axis,
                )
            else:
                ctx[op.outputs[0].key] = np.insert(
                    inp,
                    index_obj[part_index] - op.range_on_axis[0],
                    values,
                    axis=op.axis,
                )

    def __call__(self, arr, obj, values, shape):
        return self.new_tensor(
            filter_inputs([arr, obj, values]), shape=shape, order=arr.order
        )


def insert(arr, obj, values, axis=None):
    """
    Insert values along the given axis before the given indices.

    Parameters
    ----------
    arr : array like
        Input array.
    obj : int, slice or sequence of ints
        Object that defines the index or indices before which `values` is
        inserted.
    values : array_like
        Values to insert into `arr`. If the type of `values` is different
        from that of `arr`, `values` is converted to the type of `arr`.
        `values` should be shaped so that ``arr[...,obj,...] = values``
        is legal.
    axis : int, optional
        Axis along which to insert `values`.  If `axis` is None then `arr`
        is flattened first.
    Returns
    -------
    out : ndarray
        A copy of `arr` with `values` inserted.  Note that `insert`
        does not occur in-place: a new array is returned. If
        `axis` is None, `out` is a flattened array.
    See Also
    --------
    append : Append elements at the end of an array.
    concatenate : Join a sequence of arrays along an existing axis.
    delete : Delete elements from an array.
    Notes
    -----
    Note that for higher dimensional inserts `obj=0` behaves very different
    from `obj=[0]` just like `arr[:,0,:] = values` is different from
    `arr[:,[0],:] = values`.
    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.array([[1, 1], [2, 2], [3, 3]])
    >>> a.execute()
    array([[1, 1],
           [2, 2],
           [3, 3]])
    >>> mt.insert(a, 1, 5).execute()
    array([1, 5, 1, ..., 2, 3, 3])
    >>> mt.insert(a, 1, 5, axis=1).execute()
    array([[1, 5, 1],
           [2, 5, 2],
           [3, 5, 3]])
    Difference between sequence and scalars:
    >>> mt.insert(a, [1], [[1],[2],[3]], axis=1).execute()
    array([[1, 1, 1],
           [2, 2, 2],
           [3, 3, 3]])
    >>> b = a.flatten()
    >>> b.execute()
    array([1, 1, 2, 2, 3, 3])
    >>> mt.insert(b, [2, 2], [5, 6]).execute()
    array([1, 1, 5, ..., 2, 3, 3])
    >>> mt.insert(b, slice(2, 4), [5, 6]).execute()
    array([1, 1, 5, ..., 2, 3, 3])
    >>> mt.insert(b, [2, 2], [7.13, False]).execute() # type casting
    array([1, 1, 7, ..., 2, 3, 3])
    >>> x = mt.arange(8).reshape(2, 4)
    >>> idx = (1, 3)
    >>> mt.insert(x, idx, 999, axis=1).execute()
    array([[  0, 999,   1,   2, 999,   3],
           [  4, 999,   5,   6, 999,   7]])
    """
    arr = astensor(arr)
    if getattr(obj, "ndim", 0) > 1:  # pragma: no cover
        raise ValueError(
            "index array argument obj to insert must be one dimensional or scalar"
        )

    if axis is None:
        # if axis is None, array will be flatten
        arr_size = arr.size
        idx_length = calc_object_length(obj, size=arr_size)
        shape = (arr_size + idx_length,)
    else:
        validate_axis(arr.ndim, axis)
        idx_length = calc_object_length(obj, size=arr.shape[axis])
        shape = tuple(
            s + idx_length if i == axis else s for i, s in enumerate(arr.shape)
        )

    op = TensorInsert(index_obj=obj, values=values, axis=axis, dtype=arr.dtype)
    return op(arr, obj, values, shape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import AnyField
from ..utils import gen_random_seeds
from .core import TensorRandomOperandMixin, handle_array, TensorDistribution


class TensorNormal(TensorDistribution, TensorRandomOperandMixin):
    _input_fields_ = ["loc", "scale"]
    _op_type_ = OperandDef.RAND_NORMAL

    _fields_ = "loc", "scale", "size"
    loc = AnyField("loc")
    scale = AnyField("scale")
    _func_name = "normal"

    def __call__(self, loc, scale, chunk_size=None):
        return self.new_tensor([loc, scale], None, raw_chunk_size=chunk_size)


def normal(
    random_state, loc=0.0, scale=1.0, size=None, chunk_size=None, gpu=None, dtype=None
):
    r"""
    Draw random samples from a normal (Gaussian) distribution.

    The probability density function of the normal distribution, first
    derived by De Moivre and 200 years later by both Gauss and Laplace
    independently [2]_, is often called the bell curve because of
    its characteristic shape (see the example below).

    The normal distributions occurs often in nature.  For example, it
    describes the commonly occurring distribution of samples influenced
    by a large number of tiny, random disturbances, each with its own
    unique distribution [2]_.

    Parameters
    ----------
    loc : float or array_like of floats
        Mean ("centre") of the distribution.
    scale : float or array_like of floats
        Standard deviation (spread or "width") of the distribution.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``loc`` and ``scale`` are both scalars.
        Otherwise, ``mt.broadcast(loc, scale).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized normal distribution.

    See Also
    --------
    scipy.stats.norm : probability density function, distribution or
        cumulative density function, etc.

    Notes
    -----
    The probability density for the Gaussian distribution is

    .. math:: p(x) = \frac{1}{\sqrt{ 2 \pi \sigma^2 }}
                     e^{ - \frac{ (x - \mu)^2 } {2 \sigma^2} },

    where :math:`\mu` is the mean and :math:`\sigma` the standard
    deviation. The square of the standard deviation, :math:`\sigma^2`,
    is called the variance.

    The function has its peak at the mean, and its "spread" increases with
    the standard deviation (the function reaches 0.607 times its maximum at
    :math:`x + \sigma` and :math:`x - \sigma` [2]_).  This implies that
    `numpy.random.normal` is more likely to return samples lying close to
    the mean, rather than those far away.

    References
    ----------
    .. [1] Wikipedia, "Normal distribution",
           http://en.wikipedia.org/wiki/Normal_distribution
    .. [2] P. R. Peebles Jr., "Central Limit Theorem" in "Probability,
           Random Variables and Random Signal Principles", 4th ed., 2001,
           pp. 51, 51, 125.

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> mu, sigma = 0, 0.1 # mean and standard deviation
    >>> s = mt.random.normal(mu, sigma, 1000)

    Verify the mean and the variance:

    >>> (abs(mu - mt.mean(s)) < 0.01).execute()
    True

    >>> (abs(sigma - mt.std(s, ddof=1)) < 0.01).execute()
    True

    Display the histogram of the samples, along with
    the probability density function:

    >>> import matplotlib.pyplot as plt
    >>> count, bins, ignored = plt.hist(s.execute(), 30, normed=True)
    >>> plt.plot(bins, (1/(sigma * mt.sqrt(2 * mt.pi)) *
    ...                mt.exp( - (bins - mu)**2 / (2 * sigma**2) )).execute(),
    ...          linewidth=2, color='r')
    >>> plt.show()
    """
    if dtype is None:
        dtype = (
            np.random.RandomState()
            .normal(handle_array(loc), handle_array(scale), size=(0,))
            .dtype
        )
    size = random_state._handle_size(size)
    seed = gen_random_seeds(1, random_state.to_numpy())[0]
    op = TensorNormal(size=size, seed=seed, gpu=gpu, dtype=dtype)
    return op(loc, scale, chunk_size=chunk_size)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Tuple

from ..utils import lazy_import
from .core import Serializer, buffered, PickleSerializer
from .exception import ExceptionSerializer

ray = lazy_import("ray")


class RaySerializer(Serializer):
    """Return raw object to let ray do serialization."""

    @buffered
    def serial(self, obj: Any, context: Dict):
        return (obj,), [], True

    def deserial(self, serialized: Tuple, context: Dict, subs: List[Any]):
        assert not subs
        return serialized[0]


def register_ray_serializers():
    PickleSerializer.unregister(object)
    ExceptionSerializer.unregister(Exception)
    RaySerializer.register(object)
    RaySerializer.register(ray.ObjectRef)
    RaySerializer.register(ray.actor.ActorHandle)


def unregister_ray_serializers():
    RaySerializer.unregister(ray.actor.ActorHandle)
    RaySerializer.unregister(ray.ObjectRef)
    RaySerializer.unregister(object)
    PickleSerializer.register(object)
    ExceptionSerializer.register(Exception)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...serialization.serializables import BoolField
from ..entity import OutputType, register_fetch_class
from .base import Operand
from .core import TileableOperandMixin
from .fetch import FetchMixin, Fetch
from .fuse import Fuse, FuseChunkMixin


class ObjectOperand(Operand):
    pass


class ObjectOperandMixin(TileableOperandMixin):
    _output_type_ = OutputType.object

    def get_fuse_op_cls(self, obj):
        return ObjectFuseChunk


class ObjectFuseChunkMixin(FuseChunkMixin, ObjectOperandMixin):
    __slots__ = ()


class ObjectFuseChunk(ObjectFuseChunkMixin, Fuse):
    pass


class ObjectFetch(FetchMixin, ObjectOperandMixin, Fetch):
    _output_type_ = OutputType.object

    def __init__(self, **kw):
        kw.pop("output_types", None)
        kw.pop("_output_types", None)
        super().__init__(**kw)

    def _new_chunks(self, inputs, kws=None, **kw):
        if "_key" in kw and self.source_key is None:
            self.source_key = kw["_key"]
        return super()._new_chunks(inputs, kws=kws, **kw)

    def _new_tileables(self, inputs, kws=None, **kw):
        if "_key" in kw and self.source_key is None:
            self.source_key = kw["_key"]
        return super()._new_tileables(inputs, kws=kws, **kw)


register_fetch_class(OutputType.object, ObjectFetch, None)


class MergeDictOperand(ObjectOperand, ObjectOperandMixin):
    _merge = BoolField("merge")

    def __init__(self, merge=None, **kw):
        super().__init__(_merge=merge, **kw)

    @property
    def merge(self):
        return self._merge

    @classmethod
    def concat_tileable_chunks(cls, tileable):
        assert not tileable.is_coarse()

        op = cls(merge=True)
        chunk = cls(merge=True).new_chunk(tileable.chunks)
        return op.new_tileable([tileable], chunks=[chunk], nsplits=((1,),))

    @classmethod
    def execute(cls, ctx, op):
        assert op.merge
        inputs = [ctx[inp.key] for inp in op.inputs]
        ctx[op.outputs[0].key] = next(inp for inp in inputs if inp)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures as futures
import functools
import itertools
import logging
import time
from abc import ABC
from collections import namedtuple
from dataclasses import dataclass
from typing import Any, Callable, Coroutine, Dict, List, Set, Tuple, Type
from urllib.parse import urlparse

from ....oscar.profiling import ProfilingData
from ....serialization import serialize, deserialize
from ....serialization.ray import register_ray_serializers
from ....metrics import Metrics
from ....utils import lazy_import, lazy_import_on_load, implements, classproperty, Timer
from ...debug import debug_async_timeout
from ...errors import ServerClosed
from ..communication.base import Channel, ChannelType, Server, Client
from ..communication.core import register_client, register_server
from ..communication.errors import ChannelClosed
from .utils import report_event

ray = lazy_import("ray")
logger = logging.getLogger(__name__)

ChannelID = namedtuple(
    "ChannelID", ["local_address", "client_id", "channel_index", "dest_address"]
)

SERIALIZATION_TIMEOUT_MILLS = 1000
DESERIALIZATION_TIMEOUT_MILLS = 1000


def msg_to_simple_str(msg):  # pragma: no cover
    """An helper that prints message structure without generate a big str."""
    from ..message import SendMessage, _MessageBase

    if type(msg) == _ArgWrapper:
        msg = msg.message
    if isinstance(msg, SendMessage):
        return f"{str(type(msg).__name__)}(actor_ref={msg.actor_ref}, content={msg_to_simple_str(msg.content)})"
    if isinstance(msg, _MessageBase):
        return str(msg)
    if isinstance(msg, List):
        part_str = ", ".join([msg_to_simple_str(item) for item in msg[:5]])
        return f"List<{part_str}...{len(msg)}>"
    if isinstance(msg, Set):
        part_str = ", ".join([msg_to_simple_str(item) for item in list(msg)[:5]])
        return f"Set<{part_str}...{len(msg)}>"
    if isinstance(msg, Tuple):
        part_str = ", ".join([msg_to_simple_str(item) for item in msg[:5]])
        return f"Tuple<{part_str}...{len(msg)}>"
    if isinstance(msg, Dict):
        part_str = []
        it = iter(msg.items())
        try:
            while len(part_str) < 5:
                entry = next(it)
                part_str.append(
                    f"k={msg_to_simple_str(entry[0])}, v={msg_to_simple_str(entry[1])}"
                )
        except StopIteration:
            pass
        part_str = ", ".join(part_str)
        return f"Dict<{part_str}...{len(msg)}>"
    if isinstance(msg, (str, float, int, bool)):
        return "{!s:.50}".format(msg)
    return str(type(msg))


_register_ray_serializers_once = functools.lru_cache(1)(register_ray_serializers)


def _argwrapper_unpickler(serialized_message):
    _register_ray_serializers_once()
    return _ArgWrapper(deserialize(*serialized_message))


@dataclass
class _ArgWrapper:
    message: Any = None

    def __init__(self, message):
        self.message = message

    def __reduce__(self):
        _register_ray_serializers_once()
        return _argwrapper_unpickler, (serialize(self.message),)


@lazy_import_on_load(ray)
def _init_ray_metrics():
    # Note: Must init metrics before using and here initializing metrics
    # with ray backend.
    from ....metrics import init_metrics

    init_metrics("ray")
    _ray_serialize = ray.serialization.SerializationContext.serialize
    _ray_deserialize_object = ray.serialization.SerializationContext._deserialize_object
    serialized_bytes_counter = Metrics.counter(
        "mars.channel_serialized_bytes",
        "The bytes serialized by mars ray channel.",
    )
    deserialized_bytes_counter = Metrics.counter(
        "mars.channel_deserialized_bytes",
        "The bytes deserialized by mars ray channel.",
    )
    serialization_time_mills = Metrics.counter(
        "mars.channel_serialization_time_mills",
        "The time used by mars ray channel serialization.",
    )
    deserialization_time_mills = Metrics.counter(
        "mars.channel_deserialization_time_mills",
        "The time used by mars ray channel deserialization.",
    )

    def _serialize(self, value):
        if type(value) is _ArgWrapper:  # pylint: disable=unidiomatic-typecheck
            message = value.message
            with Timer() as timer:
                serialized_object = _ray_serialize(self, value)
                bytes_length = serialized_object.total_bytes
                serialized_bytes_counter.record(bytes_length)
            serialization_time_mills.record(timer.duration * 1000)
            if timer.duration * 1000 > SERIALIZATION_TIMEOUT_MILLS:  # pragma: no cover
                report_event(
                    "WARNING",
                    "SERIALIZATION_TIMEOUT",
                    f"Serialization took {timer.duration} seconds for {bytes_length} sized message {msg_to_simple_str(message)}.",
                )
            try:
                if message.profiling_context is not None:
                    task_id = message.profiling_context.task_id
                    ProfilingData[task_id, "serialization"].inc(
                        "serialize", timer.duration
                    )
            except AttributeError:  # pragma: no cover
                logger.info(
                    "Profiling serialization got error, the send "
                    "message %s may not be an instance of message",
                    type(message),
                )
        else:
            serialized_object = _ray_serialize(self, value)
        return serialized_object

    def _deserialize_object(self, data, metadata, object_ref):
        start_time = time.time()
        bytes_length = 0
        if data:
            bytes_length = len(data)
            deserialized_bytes_counter.record(bytes_length)
        value = _ray_deserialize_object(self, data, metadata, object_ref)
        duration = time.time() - start_time
        deserialization_time_mills.record(duration * 1000)
        if duration * 1000 > DESERIALIZATION_TIMEOUT_MILLS:  # pragma: no cover
            report_event(
                "WARNING",
                "DESERIALIZATION_TIMEOUT",
                f"Deserialization took {duration} seconds for "
                f"{bytes_length} sized msg {msg_to_simple_str(value)}",
            )
        if type(value) is _ArgWrapper:  # pylint: disable=unidiomatic-typecheck
            message = value.message
            try:
                if message.profiling_context is not None:
                    task_id = message.profiling_context.task_id
                    ProfilingData[task_id, "serialization"].inc(
                        "deserialize", time.time() - start_time
                    )
            except AttributeError:  # pragma: no cover
                logger.info(
                    "Profiling serialization got error, the recv "
                    "message %s may not be an instance of message",
                    type(message),
                )
        return value

    ray.serialization.SerializationContext.serialize = _serialize
    ray.serialization.SerializationContext._deserialize_object = _deserialize_object


class RayChannelException(Exception):
    def __init__(self, exc_type, exc_value: BaseException, exc_traceback):
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.exc_traceback = exc_traceback


class RayChannelBase(Channel, ABC):
    """
    Channel for communications between ray processes.
    """

    __slots__ = "_channel_index", "_channel_id", "_closed"

    name = "ray"
    _channel_index_gen = itertools.count()

    def __init__(
        self,
        local_address: str = None,
        dest_address: str = None,
        channel_index: int = None,
        channel_id: ChannelID = None,
        compression=None,
    ):
        super().__init__(
            local_address=local_address,
            dest_address=dest_address,
            compression=compression,
        )
        self._channel_index = channel_index or next(self._channel_index_gen)
        self._channel_id = channel_id or ChannelID(
            local_address, _gen_client_id(), self._channel_index, dest_address
        )
        self._closed = asyncio.Event()

    @property
    def channel_id(self) -> ChannelID:
        return self._channel_id

    @property
    @implements(Channel.type)
    def type(self) -> ChannelType:
        return ChannelType.ray

    @implements(Channel.close)
    async def close(self):
        self._closed.set()

    @property
    @implements(Channel.closed)
    def closed(self) -> bool:
        return self._closed.is_set()


class RayClientChannel(RayChannelBase):
    """
    A channel from ray driver/actor to ray actor. Use ray call reply for client channel recv.
    """

    __slots__ = "_peer_actor", "_done", "_todo"

    def __init__(
        self,
        dest_address: str = None,
        channel_index: int = None,
        channel_id: ChannelID = None,
        compression=None,
    ):
        super().__init__(None, dest_address, channel_index, channel_id, compression)
        # ray actor should be created with the address as the name.
        self._peer_actor: "ray.actor.ActorHandle" = ray.get_actor(dest_address)
        self._done = asyncio.Queue()
        self._todo = set()

    def _submit_task(self, message: Any, object_ref: "ray.ObjectRef"):
        async def handle_task(message: Any, object_ref: "ray.ObjectRef"):
            # use `%.500` to avoid print too long messages
            with debug_async_timeout(
                "ray_object_retrieval_timeout",
                "Message that client sent to actor %s is %.500s and object_ref is %s",
                self.dest_address,
                message,
                object_ref,
            ):
                try:
                    result = await object_ref
                except Exception as e:  # pragma: no cover
                    # The error ClientObjectRef can't be formatted, so
                    # we give it a string `ClientObjectRef` instead.
                    try:
                        object_ref_str = str(object_ref)
                    except Exception:
                        object_ref_str = "ClientObjectRef"
                    logger.exception(
                        "Get object %s from %s failed, got exception %s.",
                        object_ref_str,
                        self.dest_address,
                        e,
                    )
                    raise
            if isinstance(result, RayChannelException):
                raise result.exc_value.with_traceback(result.exc_traceback)
            return result.message

        def _on_completion(future):
            self._todo.remove(future)
            self._done.put_nowait(future)

        future = asyncio.ensure_future(handle_task(message, object_ref))
        future.add_done_callback(_on_completion)
        self._todo.add(future)

    @implements(Channel.send)
    async def send(self, message: Any):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed("Channel already closed, cannot send message")
        # Put ray object ref to todo queue
        task = self._peer_actor.__on_ray_recv__.remote(
            self.channel_id, _ArgWrapper(message)
        )
        self._submit_task(message, task)
        await asyncio.sleep(0)

    @implements(Channel.recv)
    async def recv(self):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed("Channel already closed, cannot recv message")
        try:
            # Wait first done.
            future = await self._done.get()
            return future.result()
        except ray.exceptions.RayActorError:
            if not self._closed.is_set():
                # raise a EOFError as the SocketChannel does
                raise EOFError("Server may be closed")
        except (RuntimeError, ServerClosed) as e:  # pragma: no cover
            if not self._closed.is_set():
                raise e


class RayServerChannel(RayChannelBase):
    """
    A channel from ray actor to ray driver/actor. Since ray actor can't call ray driver,
    we use ray call reply for server channel send. Note that there can't be multiple
    channel message sends for one received message, or else it will be taken as next
    message's reply.
    """

    __slots__ = "_in_queue", "_out_queue", "_msg_recv_counter", "_msg_sent_counter"

    def __init__(
        self,
        local_address: str = None,
        channel_index: int = None,
        channel_id: ChannelID = None,
        compression=None,
    ):
        super().__init__(local_address, None, channel_index, channel_id, compression)
        self._in_queue = asyncio.Queue()
        self._out_queue = asyncio.Queue()
        self._msg_recv_counter = 0
        self._msg_sent_counter = 0

    @implements(Channel.send)
    async def send(self, message: Any):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed("Channel already closed, cannot send message")
        # Current process is ray actor, we use ray call reply to send message to ray driver/actor.
        # Not that we can only send once for every read message in channel, otherwise
        # it will be taken as other message's reply.
        await self._out_queue.put(message)
        self._msg_sent_counter += 1
        assert (
            self._msg_sent_counter <= self._msg_recv_counter
        ), "RayServerChannel channel doesn't support send multiple replies for one message."

    @implements(Channel.recv)
    async def recv(self):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed("Channel already closed, cannot write message")
        try:
            return await self._in_queue.get()
        except RuntimeError:  # pragma: no cover
            if not self._closed.is_set():
                raise

    async def __on_ray_recv__(self, message_wrapper):
        """This method will be invoked when current process is a ray actor rather than a ray driver"""
        self._msg_recv_counter += 1
        await self._in_queue.put(message_wrapper.message)
        result_message = await self._out_queue.get()
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed("Channel already closed")
        return _ArgWrapper(result_message)

    @implements(Channel.close)
    async def close(self):
        await super().close()
        self._out_queue.put_nowait(None)


@register_server
class RayServer(Server):
    __slots__ = "_closed", "_channels", "_tasks"

    scheme = "ray"
    _server_instance = None
    _ray_actor_started = False

    def __init__(self, address, channel_handler: Callable[[Channel], Coroutine] = None):
        super().__init__(address, channel_handler)
        self._closed = asyncio.Event()
        self._channels: Dict[ChannelID, RayServerChannel] = dict()
        self._tasks: Dict[ChannelID, asyncio.Task] = dict()

    @classproperty
    @implements(Server.client_type)
    def client_type(self) -> Type["Client"]:
        return RayClient

    @property
    @implements(Server.channel_type)
    def channel_type(self) -> ChannelType:
        return ChannelType.ray

    @classmethod
    def set_ray_actor_started(cls):
        cls._ray_actor_started = True

    @classmethod
    def is_ray_actor_started(cls):
        return cls._ray_actor_started

    @staticmethod
    @implements(Server.create)
    async def create(config: Dict) -> "RayServer":
        if not RayServer.is_ray_actor_started():
            logger.warning(
                "Current process is not a ray actor, the ray server "
                "will not receive messages from clients."
            )
        assert RayServer._server_instance is None
        config = config.copy()
        address = config.pop("address")
        handle_channel = config.pop("handle_channel")
        if urlparse(address).scheme != RayServer.scheme:  # pragma: no cover
            raise ValueError(
                f"Address for RayServer "
                f'should be starts with "ray://", '
                f"got {address}"
            )
        if config:  # pragma: no cover
            raise TypeError(
                f"Creating RayServer got unexpected " f'arguments: {",".join(config)}'
            )
        server = RayServer(address, handle_channel)
        RayServer._server_instance = server
        return server

    @classmethod
    def get_instance(cls):
        return cls._server_instance

    @classmethod
    def clear(cls):
        cls._server_instance = None
        cls._ray_actor_started = False

    @implements(Server.start)
    async def start(self):
        # nothing needs to do for ray server
        pass

    @implements(Server.join)
    async def join(self, timeout=None):
        wait_coro = self._closed.wait()
        try:
            await asyncio.wait_for(wait_coro, timeout=timeout)
        except (futures.TimeoutError, asyncio.TimeoutError):  # pragma: no cover
            pass

    @implements(Server.on_connected)
    async def on_connected(self, *args, **kwargs):
        channel = args[0]
        assert isinstance(channel, RayServerChannel)
        if kwargs:  # pragma: no cover
            raise TypeError(
                f"{type(self).__name__} got unexpected "
                f'arguments: {",".join(kwargs)}'
            )
        await self.channel_handler(channel)

    @implements(Server.stop)
    async def stop(self):
        self._closed.set()
        for task in self._tasks.values():
            task.cancel()
        self._tasks = dict()
        for channel in self._channels.values():
            await channel.close()
        self._channels = dict()
        self.clear()

    @property
    @implements(Server.stopped)
    def stopped(self) -> bool:
        return self._closed.is_set()

    async def __on_ray_recv__(self, channel_id: ChannelID, message):
        if self.stopped:
            raise ServerClosed(
                f"Remote server {self.address} closed, but got message {message} "
                f"from channel {channel_id}"
            )
        channel = self._channels.get(channel_id)
        if not channel:
            _, _, peer_channel_index, peer_dest_address = channel_id
            channel = RayServerChannel(
                peer_dest_address, peer_channel_index, channel_id
            )
            self._channels[channel_id] = channel
            self._tasks[channel_id] = asyncio.create_task(self.on_connected(channel))
        return await channel.__on_ray_recv__(message)


@register_client
class RayClient(Client):
    __slots__ = ()

    scheme = RayServer.scheme

    def __init__(self, local_address: str, dest_address: str, channel: Channel):
        super().__init__(local_address, dest_address, channel)

    @staticmethod
    @implements(Client.connect)
    async def connect(
        dest_address: str, local_address: str = None, **kwargs
    ) -> "Client":
        if urlparse(dest_address).scheme != RayServer.scheme:  # pragma: no cover
            raise ValueError(
                f'Destination address should start with "ray://" '
                f"for RayClient, got {dest_address}"
            )
        client_channel = RayClientChannel(dest_address)
        client = RayClient(local_address, dest_address, client_channel)
        return client

    @implements(Client.close)
    async def close(self):
        await super().close()


def _gen_client_id():
    import uuid

    return uuid.uuid4().hex



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ... import opcodes
from ..arithmetic.core import TensorUnaryOp, TensorBinOp, TensorMultiOp
from ..array_utils import (
    np,
    cp,
    issparse,
    sparse,
    convert_order,
    as_same_device,
    device,
)


_func_name_to_special_cls = {}


def _register_special_op(cls):
    if cls._func_name is not None:
        _func_name_to_special_cls[cls._func_name] = cls
    return cls


class TensorSpecialOperandMixin:
    _op_code_ = opcodes.SPECIAL
    _func_name = None

    def __new__(cls, *args, **kwargs):
        if cls._func_name is not None:
            return object.__new__(_func_name_to_special_cls[cls._func_name])
        return super().__new__(cls, *args, **kwargs)

    @classmethod
    def _get_func(cls, xp):
        if xp is np:
            from scipy import special

            return getattr(special, cls._func_name)
        elif cp is not None and xp is cp:
            from cupyx.scipy import special

            return getattr(special, cls._func_name)
        else:
            assert xp is sparse
            return getattr(sparse, cls._func_name)


class TensorSpecialUnaryOp(TensorSpecialOperandMixin, TensorUnaryOp):
    pass


class TensorSpecialBinOp(TensorSpecialOperandMixin, TensorBinOp):
    pass


class TensorSpecialMultiOp(TensorSpecialOperandMixin, TensorMultiOp):
    @classmethod
    def _execute_gpu(cls, op, xp, *args, **kw):
        if kw.get("out") is not None:
            kw["out"] = xp.asarray(kw["out"])
        r = cls._get_func(xp)(*args, **kw)
        return convert_order(r, op.outputs[0].order.value)

    @classmethod
    def _execute_cpu(cls, op, xp, *args, **kw):
        kw["order"] = op.order
        if kw.get("out") is not None:
            kw["out"] = np.asarray(kw["out"])
        try:
            return cls._get_func(xp)(*args, **kw)
        except TypeError:
            kw.pop("order")
            r = cls._get_func(xp)(*args, **kw)
            if issparse(r):
                return r
            return convert_order(r, op.outputs[0].order.value)

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            kw = {"casting": op.casting} if op.out is not None else {}

            inputs_iter = iter(inputs)
            args = [a if np.isscalar(a) else next(inputs_iter) for a in op.args]
            if op.out is not None:
                kw["out"] = next(inputs_iter).copy()

            with np.errstate(**op.err):
                if op.is_gpu():
                    ret = cls._execute_gpu(op, xp, *args, **kw)
                else:
                    ret = cls._execute_cpu(op, xp, *args, **kw)

                if ret.dtype != op.dtype:
                    ret = ret.astype(op.dtype)
                ctx[op.outputs[0].key] = ret



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from collections import defaultdict
from operator import itemgetter
from typing import List, Dict, Union

import numpy as np

from ....core import ChunkGraph, ChunkData
from ....core.operand import Operand
from ....lib.ordered_set import OrderedSet
from ....resource import Resource
from ....typing import BandType
from ....utils import implements


class AbstractGraphAssigner(ABC):
    """
    Assign start nodes.
    """

    def __init__(
        self,
        chunk_graph: ChunkGraph,
        start_ops: List[Operand],
        band_resource: Dict[BandType, Resource],
    ):
        self._chunk_graph = chunk_graph
        self._start_ops = start_ops
        self._band_resource = band_resource

    @abstractmethod
    def assign(self, cur_assigns: Dict[str, str] = None) -> Dict[ChunkData, BandType]:
        """
        Assign start nodes to bands.

        cur_assigns : dict
            op already assigned.

        Returns
        -------
        node_to_bands : dict
            From node to band.
        """

    def get_device_band_slots(self) -> Dict[BandType, int]:
        if self._start_ops and all(
            op.gpu for op in self._start_ops
        ):  # pragma: no cover
            band_prefix = "gpu"
        else:
            band_prefix = "numa"
        return {
            band: resource.num_cpus or resource.num_gpus
            for band, resource in self._band_resource.items()
            if band[1].startswith(band_prefix)
        }


class GraphAssigner(AbstractGraphAssigner):
    def __init__(
        self,
        chunk_graph: ChunkGraph,
        start_ops: List[Operand],
        band_resource: Dict[BandType, Resource],
    ):
        super().__init__(chunk_graph, start_ops, band_resource)
        self._op_keys: OrderedSet[str] = OrderedSet(
            [start_op.key for start_op in start_ops]
        )

    def _calc_band_assign_limits(
        self, initial_count: int, occupied: Dict[BandType, int]
    ) -> Dict[BandType, int]:
        """
        Calculate limitation of number of initial operands for bands.

        Parameters
        ----------
        initial_count : int
            Number of nodes that is ready for running.
        occupied : dict
           Band to those initials that already assigned.

        Returns
        -------
        slot_assign_limits: dict
            Slot to limitation of number of initial operands.
        """
        actual_count: int = initial_count - sum(occupied.values())
        band_slots = sorted(
            self.get_device_band_slots().items(), key=itemgetter(1), reverse=True
        )
        bands: List[BandType] = [it[0] for it in band_slots]
        slots = np.asarray([it[1] for it in band_slots], dtype=np.float32)

        # remove assigned nodes from limitations
        counts = initial_count * slots / slots.sum()
        for i, band in enumerate(bands):
            counts[i] = max(0, counts[i] - occupied.get(band, 0))

        # all assigned, nothing to do
        if counts.sum() == 0:
            return {band: 0 for band in bands}

        # assign remaining nodes
        counts = (actual_count * counts / counts.sum()).astype(np.int32)
        pos = 0
        rest = actual_count - counts.sum()
        while rest > 0:
            counts[pos] += 1
            rest -= 1
            pos = (pos + 1) % len(counts)
        return dict(zip(bands, counts))

    @classmethod
    def _assign_by_bfs(
        cls,
        undirected_chunk_graph: ChunkGraph,
        start: ChunkData,
        band: BandType,
        initial_sizes: Dict[BandType, int],
        spread_limits: Dict[BandType, float],
        key_to_assign: OrderedSet[str],
        assigned_record: Dict[str, Union[str, BandType]],
    ):
        """
        Assign initial nodes using breath-first search given initial sizes and
        limitations of spread range.
        """
        if initial_sizes[band] <= 0:
            return

        assigned = 0
        spread_range = 0
        for chunk in undirected_chunk_graph.bfs(start=start, visit_predicate="all"):
            op_key = chunk.op.key
            if op_key in assigned_record:
                continue
            spread_range += 1
            # `op_key` may not be in `key_to_assign`,
            # but we need to record it to avoid iterate the node repeatedly.
            assigned_record[op_key] = band
            if op_key not in key_to_assign:
                continue
            assigned += 1
            if spread_range >= spread_limits[band] or assigned >= initial_sizes[band]:
                break
        initial_sizes[band] -= assigned

    def _build_undirected_chunk_graph(
        self, chunk_to_assign: List[ChunkData]
    ) -> ChunkGraph:
        chunk_graph = self._chunk_graph.copy()
        # remove edges for all chunk_to_assign which may contain chunks
        # that need be reassigned
        for chunk in chunk_to_assign:
            if chunk_graph.count_predecessors(chunk) > 0:
                for pred in list(chunk_graph.predecessors(chunk)):
                    chunk_graph.remove_edge(pred, chunk)
        return chunk_graph.build_undirected()

    @implements(AbstractGraphAssigner.assign)
    def assign(
        self, cur_assigns: Dict[str, BandType] = None
    ) -> Dict[ChunkData, BandType]:
        graph = self._chunk_graph
        assign_result = dict()
        cur_assigns = cur_assigns or dict()
        # assigned by expect worker
        initial_assigned_op_keys = set(cur_assigns)

        op_key_to_chunks = defaultdict(list)
        for chunk in graph:
            op_key_to_chunks[chunk.op.key].append(chunk)

        op_keys = OrderedSet(self._op_keys)
        chunk_to_assign = [
            op_key_to_chunks[op_key][0]
            for op_key in op_keys
            if op_key not in cur_assigns
        ]
        assigned_counts = defaultdict(lambda: 0)
        for band in cur_assigns.values():
            assigned_counts[band] += 1

        # build undirected graph
        undirected_chunk_graph = self._build_undirected_chunk_graph(chunk_to_assign)

        # calculate the number of chunks to be assigned to each band
        # given number of bands and existing assignments
        band_quotas = self._calc_band_assign_limits(
            len(chunk_to_assign) + sum(assigned_counts.values()), assigned_counts
        )

        # calculate expected descendant count (spread range) of
        # every band and subtract assigned number from it
        average_spread_range = len(graph) * 1.0 / len(self.get_device_band_slots())
        spread_ranges = defaultdict(lambda: average_spread_range)
        # assign from other chunks to be assigned
        # TODO: sort by what?
        sorted_candidates = chunk_to_assign.copy()
        while max(band_quotas.values()):
            band = max(band_quotas, key=lambda k: band_quotas[k])
            cur = sorted_candidates.pop()
            while cur.op.key in cur_assigns:
                cur = sorted_candidates.pop()
            self._assign_by_bfs(
                undirected_chunk_graph,
                cur,
                band,
                band_quotas,
                spread_ranges,
                op_keys,
                cur_assigns,
            )

        key_to_assign = {n.op.key for n in chunk_to_assign} | initial_assigned_op_keys
        for op_key, band in cur_assigns.items():
            if op_key in key_to_assign:
                for chunk in op_key_to_chunks[op_key]:
                    assign_result[chunk] = band

        return assign_result



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import pickle
from collections import OrderedDict, defaultdict

import numpy as np

from .... import opcodes as OperandDef
from ....core import OutputType
from ....core.context import get_context
from ....core.operand import MergeDictOperand
from ....serialization.serializables import FieldTypes, DictField, KeyField, ListField
from ....utils import ensure_own_data
from .dmatrix import ToDMatrix, to_dmatrix
from .start_tracker import StartTracker

logger = logging.getLogger(__name__)


def _on_serialize_evals(evals_val):
    if evals_val is None:
        return None
    return [list(x) for x in evals_val]


class XGBTrain(MergeDictOperand):
    _op_type_ = OperandDef.XGBOOST_TRAIN

    params = DictField("params", key_type=FieldTypes.string, default=None)
    dtrain = KeyField("dtrain", default=None)
    evals = ListField("evals", on_serialize=_on_serialize_evals, default=None)
    kwargs = DictField("kwargs", key_type=FieldTypes.string, default=None)
    tracker = KeyField("tracker", default=None)

    def __init__(self, gpu=None, **kw):
        super().__init__(gpu=gpu, **kw)
        if self.output_types is None:
            self.output_types = [OutputType.object]

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self.dtrain = self._inputs[0]
        rest = self._inputs[1:]
        if self.tracker is not None:
            self.tracker = self._inputs[-1]
            rest = rest[:-1]
        if self.evals is not None:
            evals_dict = OrderedDict(self.evals)
            new_evals_dict = OrderedDict()
            for new_key, val in zip(rest, evals_dict.values()):
                new_evals_dict[new_key] = val
            self.evals = list(new_evals_dict.items())

    def __call__(self):
        inputs = [self.dtrain]
        if self.evals is not None:
            inputs.extend(e[0] for e in self.evals)
        return self.new_tileable(inputs)

    @staticmethod
    def _get_dmatrix_chunks_workers(ctx, dmatrix):
        # dmatrix_chunk.inputs is concat, and concat's input is the coallocated chunks
        metas = ctx.get_chunks_meta(
            [c.inputs[0].inputs[0].key for c in dmatrix.chunks], fields=["bands"]
        )
        return [m["bands"][0][0] for m in metas]

    @classmethod
    def tile(cls, op: "XGBTrain"):
        ctx = get_context()

        inp = op.inputs[0]
        in_chunks = inp.chunks
        workers = cls._get_dmatrix_chunks_workers(ctx, inp)
        worker_to_in_chunks = dict(zip(workers, in_chunks))
        n_chunk = len(in_chunks)
        out_chunks = []
        worker_to_evals = defaultdict(dict)
        if op.evals is not None:
            for dm, ev in op.evals:
                ev_workers = cls._get_dmatrix_chunks_workers(ctx, dm)
                for ev_worker, ev_chunk in zip(ev_workers, dm.chunks):
                    worker_to_evals[ev_worker][ev] = ev_chunk

        all_workers = set(workers)
        all_workers.update(worker_to_evals)

        i = itertools.count(n_chunk)
        tracker_chunk = StartTracker(
            n_workers=len(all_workers), pure_depends=[True] * n_chunk
        ).new_chunk(in_chunks, shape=())
        for worker in all_workers:
            chunk_op = op.copy().reset_key()
            chunk_op.expect_worker = worker
            chunk_op.tracker = tracker_chunk
            if worker in worker_to_in_chunks:
                in_chunk = worker_to_in_chunks[worker]
            else:
                in_chunk_op = ToDMatrix(
                    data=None,
                    label=None,
                    weight=None,
                    base_margin=None,
                    missing=inp.op.missing,
                    feature_names=inp.op.feature_names,
                    feature_types=inp.op.feature_types,
                    _output_types=inp.op.output_types,
                )
                params = inp.params.copy()
                params["index"] = (next(i), 0)
                params["shape"] = (0, inp.shape[1])
                in_chunk = in_chunk_op.new_chunk(None, kws=[params])
            chunk_evals = []
            for dm, ev in op.evals:
                try:
                    chunk_evals.append((worker_to_evals[worker][ev], ev))
                except KeyError:
                    # create a new eval chunk
                    eval_chunk_op = ToDMatrix(
                        data=None,
                        label=None,
                        weight=None,
                        base_margin=None,
                        missing=dm.op.missing,
                        feature_names=dm.op.feature_names,
                        feature_types=dm.op.feature_types,
                        _output_types=dm.op.output_types,
                    )
                    params = dm.params.copy()
                    params["index"] = (0, 0)
                    params["shape"] = (0, dm.shape[1])
                    eval_chunk = eval_chunk_op.new_chunk(None, kws=[params])
                    chunk_evals.append((eval_chunk, ev))
            chunk_op.evals = chunk_evals
            input_chunks = (
                [in_chunk] + [pair[0] for pair in chunk_evals] + [tracker_chunk]
            )
            out_chunk = chunk_op.new_chunk(
                input_chunks, shape=(np.nan,), index=in_chunk.index[:1]
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs, chunks=out_chunks, nsplits=((np.nan for _ in out_chunks),)
        )

    @classmethod
    def execute(cls, ctx, op: "XGBTrain"):
        if op.merge:
            return super().execute(ctx, op)

        from xgboost import train, rabit

        params = op.params.copy()

        n_threads = 0
        if op.tracker is None:
            # non distributed
            ctx_n_threads = -1
        else:
            # distributed
            ctx_n_threads = ctx.get_slots()

        # fix parallelism on nodes
        for p in ["nthread", "n_jobs"]:
            if (
                params.get(p, None) is not None
                and params.get(p, ctx_n_threads) != ctx_n_threads
            ):  # pragma: no cover
                logger.info("Overriding `nthreads` defined in Mars worker.")
                n_threads = params[p]
                break
        if n_threads == 0 or n_threads is None:  # pragma: no branch
            n_threads = ctx_n_threads
        params.update({"nthread": n_threads, "n_jobs": n_threads})

        dtrain = ToDMatrix.get_xgb_dmatrix(
            ensure_own_data(ctx[op.dtrain.key]), nthread=n_threads
        )
        evals = tuple()
        if op.evals is not None:
            eval_dmatrices = [
                ToDMatrix.get_xgb_dmatrix(
                    ensure_own_data(ctx[t[0].key]), nthread=n_threads
                )
                for t in op.evals
            ]
            evals = tuple((m, ev[1]) for m, ev in zip(eval_dmatrices, op.evals))

        if op.tracker is None:
            # non distributed
            local_history = dict()
            kwargs = dict() if op.kwargs is None else op.kwargs
            bst = train(
                params, dtrain, evals=evals, evals_result=local_history, **kwargs
            )
            ctx[op.outputs[0].key] = {
                "booster": pickle.dumps(bst),
                "history": local_history,
            }
        else:
            # distributed
            logger.debug("Distributed train params: %r", params)

            rabit_args = ctx[op.tracker.key]
            rabit.init(
                [
                    arg.tobytes() if isinstance(arg, memoryview) else arg
                    for arg in rabit_args
                ]
            )
            try:
                logger.debug(
                    "Start to train data, train size: %s, evals sizes: %s",
                    dtrain.num_row(),
                    [ev[0].num_row() for ev in evals],
                )
                local_history = dict()
                bst = train(
                    params, dtrain, evals=evals, evals_result=local_history, **op.kwargs
                )
                ret = {"booster": pickle.dumps(bst), "history": local_history}
                if rabit.get_rank() != 0:
                    ret = {}
                ctx[op.outputs[0].key] = ret
            finally:
                rabit.finalize()


def train(params, dtrain, evals=(), **kwargs):
    """
    Train XGBoost model in Mars manner.

    Parameters
    ----------
    Parameters are the same as `xgboost.train`.

    Returns
    -------
    results: Booster
    """

    evals_result = kwargs.pop("evals_result", dict())
    session = kwargs.pop("session", None)
    run_kwargs = kwargs.pop("run_kwargs", dict())

    processed_evals = []
    if evals:
        for eval_dmatrix, name in evals:
            if not isinstance(name, str):
                raise TypeError("evals must a list of pairs (DMatrix, string)")
            if hasattr(eval_dmatrix, "op") and isinstance(eval_dmatrix.op, ToDMatrix):
                processed_evals.append((eval_dmatrix, name))
            else:
                processed_evals.append((to_dmatrix(eval_dmatrix), name))

    op = XGBTrain(params=params, dtrain=dtrain, evals=processed_evals, kwargs=kwargs)
    t = op()
    ret = t.execute(session=session, **run_kwargs).fetch(session=session)
    evals_result.update(ret["history"])
    bst = pickle.loads(ret["booster"])
    num_class = params.get("num_class")
    if num_class:
        bst.set_attr(num_class=str(num_class))
    return bst



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import recursive_tile
from ...serialization.serializables import AnyField, StringField
from ...utils import calc_nsplits
from ..operands import DataFrameOperand, DataFrameOperandMixin, OutputType
from ..utils import build_empty_df, parse_index, standardize_range_index


class DataFrameMelt(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.MELT

    _id_vars = AnyField("id_vars")
    _value_vars = AnyField("value_vars")
    _var_name = StringField("var_name")
    _value_name = StringField("value_name")
    _col_level = AnyField("col_level")

    def __init__(
        self,
        id_vars=None,
        value_vars=None,
        var_name=None,
        value_name=None,
        col_level=None,
        **kw
    ):
        super().__init__(
            _id_vars=id_vars,
            _value_vars=value_vars,
            _var_name=var_name,
            _value_name=value_name,
            _col_level=col_level,
            **kw
        )

    @property
    def id_vars(self):
        return self._id_vars

    @property
    def value_vars(self):
        return self._value_vars

    @property
    def var_name(self):
        return self._var_name

    @property
    def value_name(self):
        return self._value_name

    @property
    def col_level(self):
        return self._col_level

    def __call__(self, df):
        empty_result = build_empty_df(df.dtypes).melt(
            id_vars=self.id_vars,
            value_vars=self.value_vars,
            var_name=self.var_name,
            value_name=self.value_name,
            col_level=self.col_level,
        )
        self._output_types = [OutputType.dataframe]
        return self.new_tileable(
            [df],
            shape=(np.nan, len(empty_result.columns)),
            dtypes=empty_result.dtypes,
            index_value=parse_index(pd.RangeIndex(-1), df.key, df.index_value.key),
            columns_value=parse_index(empty_result.columns, store_data=True),
        )

    @classmethod
    def tile(cls, op: "DataFrameMelt"):
        inp = op.inputs[0]
        out = op.outputs[0]

        inp = yield from recursive_tile(inp.rechunk({1: (inp.shape[1],)}))

        chunks = []
        for c in inp.chunks:
            new_op = op.copy().reset_key()
            chunks.append(
                new_op.new_chunk(
                    [c],
                    index=c.index,
                    shape=(np.nan, out.shape[1]),
                    dtypes=out.dtypes,
                    index_value=parse_index(
                        pd.RangeIndex(-1), c.key, c.index_value.key
                    ),
                    columns_value=out.columns_value,
                )
            )

        yield chunks
        chunks = standardize_range_index(chunks)
        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            [inp],
            chunks=chunks,
            nsplits=calc_nsplits({c.index: c.shape for c in chunks}),
            **out.params
        )

    @classmethod
    def execute(cls, ctx, op: "DataFrameMelt"):
        in_data = ctx[op.inputs[0].key]
        ctx[op.outputs[0].key] = in_data.melt(
            id_vars=op.id_vars,
            value_vars=op.value_vars,
            var_name=op.var_name,
            value_name=op.value_name,
            col_level=op.col_level,
        )


def melt(
    frame,
    id_vars=None,
    value_vars=None,
    var_name=None,
    value_name="value",
    col_level=None,
):
    """
    Unpivot a DataFrame from wide to long format, optionally leaving identifiers set.

    This function is useful to massage a DataFrame into a format where one
    or more columns are identifier variables (`id_vars`), while all other
    columns, considered measured variables (`value_vars`), are "unpivoted" to
    the row axis, leaving just two non-identifier columns, 'variable' and
    'value'.
    .. versionadded:: 0.20.0

    Parameters
    ----------
    id_vars : tuple, list, or ndarray, optional
        Column(s) to use as identifier variables.
    value_vars : tuple, list, or ndarray, optional
        Column(s) to unpivot. If not specified, uses all columns that
        are not set as `id_vars`.
    var_name : scalar
        Name to use for the 'variable' column. If None it uses
        ``frame.columns.name`` or 'variable'.
    value_name : scalar, default 'value'
        Name to use for the 'value' column.
    col_level : int or str, optional
        If columns are a MultiIndex then use this level to melt.

    Returns
    -------
    DataFrame
        Unpivoted DataFrame.

    See Also
    --------
    melt
    pivot_table
    DataFrame.pivot
    Series.explode

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'A': {0: 'a', 1: 'b', 2: 'c'},
    ...                    'B': {0: 1, 1: 3, 2: 5},
    ...                    'C': {0: 2, 1: 4, 2: 6}})
    >>> df.execute()
       A  B  C
    0  a  1  2
    1  b  3  4
    2  c  5  6

    >>> df.melt(id_vars=['A'], value_vars=['B']).execute()
       A variable  value
    0  a        B      1
    1  b        B      3
    2  c        B      5

    >>> df.melt(id_vars=['A'], value_vars=['B', 'C']).execute()
       A variable  value
    0  a        B      1
    1  b        B      3
    2  c        B      5
    3  a        C      2
    4  b        C      4
    5  c        C      6

    The names of 'variable' and 'value' columns can be customized:

    >>> df.melt(id_vars=['A'], value_vars=['B'],
    ...         var_name='myVarname', value_name='myValname').execute()
       A myVarname  myValname
    0  a         B          1
    1  b         B          3
    2  c         B          5

    If you have multi-index columns:

    >>> df = md.DataFrame({('A', 'D'): {0: 'a', 1: 'b', 2: 'c'},
    ...                    ('B', 'E'): {0: 1, 1: 3, 2: 5},
    ...                    ('C', 'F'): {0: 2, 1: 4, 2: 6}})
    >>> df.execute()
       A  B  C
       D  E  F
    0  a  1  2
    1  b  3  4
    2  c  5  6

    >>> df.melt(col_level=0, id_vars=['A'], value_vars=['B']).execute()
       A variable  value
    0  a        B      1
    1  b        B      3
    2  c        B      5

    >>> df.melt(id_vars=[('A', 'D')], value_vars=[('B', 'E')]).execute()
      (A, D) variable_0 variable_1  value
    0      a          B          E      1
    1      b          B          E      3
    2      c          B          E      5
    """
    op = DataFrameMelt(
        id_vars=id_vars,
        value_vars=value_vars,
        var_name=var_name,
        value_name=value_name,
        col_level=col_level,
    )
    return op(frame)
